"""
Categories and Kinds are special enumerations that classify the feedback
conditions and responses, respectively.
"""


class FeedbackCategory:
    """
    An Enumeration of feedback condition categories.
    These categories represent distinct types of feedback conditions based on their
    presence within the students' submission.
    Notice that these explain the condition, not the feedback response (which would
    fall under Kind).

    Some are contextualized to instruction ("mistakes", "specification", "instructor")
    and some are generic ("syntax", "runtime", "algorithmic"). One category is also
    available for errors identified by the student.
    """

    #: Grammatical and typographical errors that prevent parsing.
    SYNTAX = 'syntax'
    #: Execution errors triggered during runtime by an invalid Python operation
    RUNTIME = 'runtime'
    #: Errors that do not prevent functioning code but are generically wrong.
    ALGORITHMIC = 'algorithmic'
    #: Errors that do not prevent functioning code but are specifically wrong.
    MISTAKES = 'mistakes'
    #: Errors suggested because the code failed to meet specified behavior.
    SPECIFICATION = 'specification'
    #: Errors marked by the instructor in a one-off fashion.
    INSTRUCTOR = 'instructor'
    #: Errors marked by the students' own code, such as failing test cases.
    STUDENT = 'student'
    #: Stylistic errors that do not prevent correct behavior but are otherwise undesirable
    STYLE = 'style'
    #: Errors caused by the Pedal grading infrastructure or the surrounding infrastructure.
    SYSTEM = 'system'
    #: A category for feedback that are not actually errors, but is positive information.
    POSITIVE = 'positive'
    #: A special category recognizing a completed submission
    COMPLETE = 'complete'
    #: A category for feedback that are not actually errors, but is neutral.
    INSTRUCTIONS = 'instructions'
    #: A category for unknown feedback. Ideally, never used.
    UNKNOWN = 'uncategorized'

    ALIASES = {
        'parser': SYNTAX,
        'verifier': SYNTAX,
        'instructor': INSTRUCTOR,
        'analyzer': ALGORITHMIC
    }


class FeedbackKind:
    """
    An enumeration of the possible kinds of feedback responses, based on their pedagogical role.
    Valence can vary between specific instances of a kind of feedback, but some tend to
    have a specific valence.

    Attributes:
        MISCONCEPTION (str): A description of the misconception that
            is believed to be in the student's mind, or perhaps the relevant concept from the
            material that should be associated with this. ("Variables must be initialized before they are used").
        MISTAKE (str): A description of the error or bug that the student has created ("NameError on line 5: sum
            has not been defined").
        HINT (str): A suggestion for what the student can do ("Initialize the sum variable on line 2").
        CONSTRAINT (str): A description of the task requirements or task type that the student has violated
            ("You used a for loop, but this question expected you to use recursion.").
        METACOGNITIVE (str): A suggestion for more regulative strategies ("You have been working for 5
            hours, perhaps it is time to take a break?").
    """
    # Negative
    MISCONCEPTION: str = "Misconception"
    MISTAKE = "Mistake"
    HINT = "Hint"
    CONSTRAINT = "Constraint"
    METACOGNITIVE = "Metacognitive"
    REINFORCEMENT = "Reinforcement"
    # Positive
    COMPLIMENT = "Compliment"
    ENCOURAGEMENT = "Encouragement"
    # Variable
    RESULT = "Result"
    PERFORMANCE = "Performance"
    # Neutral
    INSTRUCTIONAL = "Instructional"
    META = "Meta"


class FeedbackStatus:
    """ Enumeration of feedback status outcomes. When you create a piece of
    feedback, it will be either active or inactive depeneding on whether its
    condition was met. Alternatively, it is also possible that it triggered
    an exception. It may also be delayed, indicating that it has not yet been
    checked. """
    INACTIVE = "inactive"
    ERROR = "error"
    ACTIVE = "active"
    DELAYED = "delayed"



"""
A Tool for verifying source code, chunking up source code into multiple parts, and substituting code within
the submission.

- Requires: None
- Optional: None
- Category: Syntax

Code Substitution may seem esoteric, but its what powers the code chunking. The submission is literally
modified such that the current submission is replaced by the current file.

"""

from pedal.source.constants import TOOL_NAME
from pedal.source.sections import (separate_into_sections, check_section_exists,
                                   stop_sections, next_section)
from pedal.source.source import (verify, verify_section, set_source, set_source_file, get_program)


__all__ = ['TOOL_NAME',
           'set_source', 'verify', 'set_source_file', 'get_program',
           'separate_into_sections',
           'check_section_exists', 'next_section', 'verify_section',
           ]



"""
Mocked functions that can be used to prevent malicious or accidental `eval`
behavior.
"""
import re
import types

from pedal.sandbox.exceptions import (SandboxNoMoreInputsException,
                                      SandboxPreventModule)
from pedal.utilities.system import IS_PYTHON_36


def do_nothing(*args, **kwargs):
    """ A function that does absolutely nothing. """


def _disabled_compile(source, filename, mode, flags=0, dont_inherit=False):
    """
    A version of the built-in `compile` method that fails with a runtime
    error.
    """
    raise RuntimeError("You are not allowed to call 'compile'.")


def _disabled_eval(object, globals=globals(), locals=None):
    """
    A version of the built-in `eval` method that fails with a runtime
    error.
    """
    raise RuntimeError("You are not allowed to call 'eval'.")


# -------------------------------------------------------------


def _disabled_exec(object, globals=globals(), locals=None):
    """
    A version of the built-in `exec` method that fails with a runtime
    error.
    """
    raise RuntimeError("You are not allowed to call 'exec'.")


# -------------------------------------------------------------


def _disabled_globals():
    """
    A version of the built-in `globals` method that fails with a runtime
    error.
    """
    raise RuntimeError("You are not allowed to call 'globals'.")


class FunctionNotAllowed(Exception):
    """ Exception created when a function that is not allowed is used. """


def disabled_builtin(name):
    """

    Args:
        name:

    Returns:

    """
    def _disabled_version(*args, **kwargs):
        raise FunctionNotAllowed("You are not allowed to call '{}'.".format(name))
    return _disabled_version


_OPEN_FORBIDDEN_NAMES = re.compile(r"(^[./])|(\.py$)")
_OPEN_FORBIDDEN_MODES = re.compile(r"[wa+]")

# TODO: Turn this into a function that lets us more elegantly specify valid and
# invalid filenames/paths


def _restricted_open(name, mode='r', buffering=-1):
    if _OPEN_FORBIDDEN_NAMES.search(name):
        raise RuntimeError("The filename you passed to 'open' is restricted.")
    elif _OPEN_FORBIDDEN_MODES.search(mode):
        raise RuntimeError("You are not allowed to 'open' files for writing.")
    else:
        return ORIGINAL_BUILTINS['open'](name, mode, buffering)

# TODO: Allow this to be flexible


def _restricted_import(name, globals=None, locals=None, fromlist=(), level=0):
    if name == 'pedal' or name.startswith('pedal.'):
        raise RuntimeError("You cannot import pedal!")
    return ORIGINAL_BUILTINS['__import__'](name, globals, locals, fromlist, level)


try:
    __builtins__
except NameError:
    _default_builtins = {'globals': globals,
                         'locals': locals,
                         'open': open,
                         'input': input,
                         '__import__': __import__}
else:
    if isinstance(__builtins__, types.ModuleType):
        _default_builtins = __builtins__.__dict__
    else:
        _default_builtins = __builtins__

ORIGINAL_BUILTINS = {
    'globals': _default_builtins['globals'],
    'locals': _default_builtins['locals'],
    'open': _default_builtins['open'],
    'input': _default_builtins['input'],
    'exec': _default_builtins.get('exec', _disabled_exec),
    'eval': _default_builtins.get('eval', _disabled_eval),
    'compile': _default_builtins.get('compile', _disabled_compile),
    '__import__': _default_builtins['__import__']
}


def make_inputs(input_list, repeat=None):
    """
    Helper function for creating mock user input.

    Params:
        input_list (list of str): The list of inputs to be returned
    Returns:
        function (str=>str): The mock input function that is returned, which
                             will return the next element of input_list each
                             time it is called.
    """
    generator = iter(input_list)

    def mock_input(prompt=''):
        """

        Args:
            prompt:

        Returns:

        """
        print(prompt)
        try:
            return next(generator)
        except StopIteration as SI:
            if repeat is None:
                # TODO: Make this a custom exception
                raise SandboxNoMoreInputsException("User had no more input to give.")
            else:
                return repeat

    return mock_input


def create_module(module_name):
    """
    Creates empty ModuleTypes based on the ``module_name``. Correctly
    creates parent modules for submodules as needed to fill in the path.

    Args:
        module_name: A dot-separated string of modules, as if for ``import``.

    Returns:
        :py:class:`types.ModuleType`: The root module created.
        dict[str, :py:class:`types.ModuleType`]: A dictionary of newly created
            modules, including the root and the actual target.
        :py:class:`types.ModuleType`: The target module created.
    """
    submodule_names = module_name.split(".")
    modules = {}
    root = types.ModuleType(submodule_names[0])
    modules[submodule_names[0]] = root
    reconstructed_path = submodule_names[0]
    for submodule_name in submodule_names[1:]:
        reconstructed_path += "." + submodule_name
        new_submodule = types.ModuleType(reconstructed_path)
        setattr(root, submodule_name, new_submodule)
        modules[reconstructed_path] = new_submodule
    return root, modules, modules[module_name]


class MockModule:
    def _generate_patches(self):
        return {k: v for k, v in vars(self).items()
                if not k.startswith('_')}

    def add_to_module(self, module):
        for name, value in self._generate_patches().items():
            setattr(module, name, value)


class MockDictModule(MockModule):
    def __init__(self, fields):
        self.fields = fields

    def _generate_patches(self):
        return self.fields


class BlockedModule(MockModule):
    def __init__(self, name):
        self.module_name = name

    def _generate_patches(self):
        return {'__getattr__': self.getter}

    def getter(self, key):
        " If anything asks, we prevent the module. Except for __file__. "
        # Needed to support coverage - it's okay to ask who I am.
        if key == '__file__':
            return self.module_name
        else:
            self.prevent_module()

    def prevent_module(self, *args, **kwargs):
        """

        Args:
            **kwargs:
        """
        raise SandboxPreventModule(f"You cannot import {self.module_name} from student code.")


class MockPedal(BlockedModule):
    """ TODO: Deprecated? """
    module_name = "pedal"


class MockTurtle(MockModule):
    """
    Mock Turtle Module that can be used to trace turtle calls.

    Attributes:
        calls (list of dict): The traced list of calls
        # TODO: it'd be awesome to have a way to construct a representation
        #       of the drawing result that we could autograde!
    """
    module_name = 'turtle'
    if IS_PYTHON_36:
        FIELDS = ["Canvas", "Pen", "RawPen", "RawTurtle", "Screen",
                  "ScrolledCanvas", "Shape", "TK", "TNavigator", "TPen", "Tbuffer",
                  "Terminator", "Turtle", "TurtleGraphicsError", "TurtleScreen",
                  "TurtleScreenBase", "Vec2D", "addshape", "back", "backward",
                  "begin_fill", "begin_poly", "bgcolor", "bgpic", "bk", "bye",
                  "circle", "clear", "clearscreen", "clearstamp", "clearstamps",
                  "clone", "color", "colormode", "config_dict", "deepcopy",
                  "degrees", "delay", "distance", "done", "dot", "down",
                  "end_fill", "end_poly", "exitonclick", "fd", "fillcolor",
                  "filling", "forward", "get_poly", "get_shapepoly", "getcanvas",
                  "getmethparlist", "getpen", "getscreen", "getshapes", "getturtle",
                  "goto", "heading", "hideturtle", "home", "ht", "inspect",
                  "isdown", "isfile", "isvisible", "join", "left", "listen", "lt",
                  "mainloop", "math", "mode", "numinput", "onclick", "ondrag",
                  "onkey", "onkeypress", "onkeyrelease", "onrelease",
                  "onscreenclick", "ontimer", "pd", "pen", "pencolor", "pendown",
                  "pensize", "penup", "pos", "position", "pu", "radians",
                  "read_docstrings", "readconfig", "register_shape", "reset",
                  "resetscreen", "resizemode", "right", "rt", "screensize",
                  "seth", "setheading", "setpos", "setposition", "settiltangle",
                  "setundobuffer", "setup", "setworldcoordinates", "setx", "sety",
                  "shape", "shapesize", "shapetransform", "shearfactor",
                  "showturtle", "simpledialog", "speed", "split", "st", "stamp",
                  "sys", "textinput", "tilt", "tiltangle", "time", "title",
                  "towards", "tracer", "turtles", "turtlesize", "types", "undo",
                  "undobufferentries", "up", "update", "width", "window_height",
                  "window_width", "write", "write_docstringdict", "xcor", "ycor"]

    def __init__(self):
        self.calls = []

    def _reset_turtles(self):
        self.calls = []
        
    def _generate_patches(self):
        if IS_PYTHON_36:
            # Generate patches manually
            return {name: self.getter(name) for name in self.FIELDS}
        else:
            return {'__getattr__': self.getter}

    def getter(self, key):
        " If anything asks, we prevent the module. Except for __file__. "
        # Needed to support coverage - it's okay to ask who I am.
        if key == '__file__':
            return self.module_name

        def _fake_call_wrapper(*args, **kwargs):
            self.calls.append((key, args, kwargs))
            return self._fake_call(args, kwargs)
        return _fake_call_wrapper
        
    def _fake_call(self, *args, **kwargs):
        return 0


class MockPlt(MockModule):
    """
    Mock MatPlotLib library that can be used to capture plot data.

    Attributes:
        plots (list of dict): The internal list of plot dictionaries.
    """

    def __init__(self):
        super().__init__()
        self._reset_plots()

    def show(self, **kwargs):
        """ Renders the plot """
        self.plots.append(self.active_plot)
        self._reset_plot()

    def unshown_plots(self):
        """ Checks for plots that are not yet shown. """
        return self.active_plot['data']

    def __repr__(self):
        return repr(self.plots)

    def __str__(self):
        return str(self.plots)

    def _reset_plots(self):
        self.plots = []
        self._reset_plot()

    def _reset_plot(self):
        self.active_plot = {'data': [],
                            'xlabel': None, 'ylabel': None,
                            'title': None, 'legend': False}

    def hist(self, data, **kwargs):
        """ Make a histogram """
        label = kwargs.get('label', None)
        self.active_plot['data'].append({'type': 'hist', 'values': data,
                                         'label': label})

    def plot(self, xs, ys=None, **kwargs):
        """ Make a line plot """
        label = kwargs.get('label', None)
        if ys is None:
            self.active_plot['data'].append({'type': 'line',
                                             'x': list(range(len(xs))),
                                             'y': xs, 'label': label})
        else:
            self.active_plot['data'].append({'type': 'line', 'x': xs,
                                             'y': ys, 'label': label})

    def scatter(self, xs, ys, **kwargs):
        """ Make a scatter plot """
        label = kwargs.get('label', None)
        self.active_plot['data'].append({'type': 'scatter', 'x': xs,
                                         'y': ys, 'label': label})

    def xlabel(self, label, **kwargs):
        """ Label the x-axis """
        self.active_plot['xlabel'] = label

    def title(self, label, **kwargs):
        """ Make the title """
        self.active_plot['title'] = label

    def suptitle(self, label, **kwargs):
        """ Make the super title """
        self.title(label, **kwargs)

    def ylabel(self, label, **kwargs):
        """ Label the Y-axis """
        self.active_plot['ylabel'] = label

    def legend(self, **kwargs):
        """ Show the legend """
        self.active_plot['legend'] = True

    def _generate_patches(self):
        def dummy(*args, **kwargs):
            """ This function does nothing. """

        return dict(hist=self.hist, plot=self.plot,
                    scatter=self.scatter, show=self.show,
                    xlabel=self.xlabel, ylabel=self.ylabel,
                    title=self.title, legend=self.legend,
                    xticks=dummy, yticks=dummy,
                    autoscale=dummy, axhline=dummy,
                    axhspan=dummy, axvline=dummy,
                    axvspan=dummy, clf=dummy,
                    cla=dummy, close=dummy,
                    figlegend=dummy, figimage=dummy,
                    suptitle=self.suptitle, text=dummy,
                    tick_params=dummy, ticklabel_format=dummy,
                    tight_layout=dummy, xkcd=dummy,
                    xlim=dummy, ylim=dummy,
                    xscale=dummy, yscale=dummy)



from pedal.core.feedback import Feedback
from pedal.core.commands import set_success
from pedal.core.scoring import Score, combine_scores


def parse_feedback(feedback):
    """

    Args:
        feedback:

    Returns:

    """
    message = feedback.message
    title = feedback.title or feedback.label
    return feedback.correct, feedback.score, message, title, feedback.fields


class FinalFeedback:
    """
    Internal class used for organizing the feedback into one place.
    Dumpable into a simple dictionary.
    """

    DEFAULT_NO_FEEDBACK_TITLE = "No Errors"
    DEFAULT_NO_FEEDBACK_MESSAGE = "No errors reported."

    def __init__(self, success=None, score=None, category=None, label=None, title=None,
                 message=None, data=None, hide_correctness=None,
                 suppressions=None, suppressed_labels=None):
        self.success = success
        self.score = score
        self._scores = []
        self.category = category
        self.label = label
        self.title = title
        self.message = message
        self.data = data
        self.hide_correctness = hide_correctness
        self.positives = []
        self.instructions = []
        self.systems = []
        self.suppressions = suppressions
        self.suppressed_labels = suppressed_labels

    def merge(self, feedback):
        category = feedback.category.lower()
        if category in self.suppressions:
            if True in self.suppressions[category]:
                return
            elif feedback.label.lower() in self.suppressions[category]:
                return
        if feedback.label in self.suppressed_labels:
            return
        success, partial, message, title, data = parse_feedback(feedback)
        if feedback and feedback.category == Feedback.CATEGORIES.SYSTEM:
            self.systems.append(feedback)
        if not feedback.unscored and feedback.score is not None:
            # Triggered Negative leads to opposite behavior for operator
            # Also untriggered positive feedback
            invert_logic = ((feedback.valence != feedback.NEGATIVE_VALENCE) == (not feedback))
            inversion = "!" if invert_logic else ""
            self._scores.append(f"{inversion}{partial}")
        if not feedback and feedback.else_message:
            self.positives.append(feedback)
            return
        if not feedback or feedback.muted:
            return
        if feedback.kind == Feedback.KINDS.COMPLIMENT:
            self.positives.append(feedback)
            return
        if feedback.kind == Feedback.KINDS.INSTRUCTIONAL:
            self.instructions.append(feedback)
        self.success = success and self.success
        if message is not None and self.message is None:
            self.message = message
            self.title = title
            self.category = feedback.category
            self.label = feedback.label
            self.data = data

    def finalize(self):
        if self.message is None:
            self.title = self.DEFAULT_NO_FEEDBACK_TITLE
            self.message = self.DEFAULT_NO_FEEDBACK_MESSAGE
        self.hide_correctness = self.suppressions.get('success', False)
        if (not self.hide_correctness and
                self.label == 'set_success_no_errors' and
                self.category == Feedback.CATEGORIES.COMPLETE):
            # TODO: Promote to be its own atomic feedback function
            self.title = set_success.title
            self.message = set_success.message_template
            self.score = 1
            self.success = True
        else:
            self.score = combine_scores(self._scores)
        self.success = bool(self.success)
        return self

    def __str__(self) -> str:
        return "FinalFeedback({label!r}, {title!r}, {message!r})".format(label=self.label,
                                                                         title=self.title,
                                                                         message=self.message[:50])

    def for_console(self) -> str:
        return "{label}\n{score}\n{title}\n{message}".format(label=self.label,
                                                             score=self.score,
                                                             title=self.title,
                                                             message=self.message)

    def to_json(self) -> dict:
        """

        Returns:

        """
        return {
            'success': self.success,
            'score': self.score,
            "scores": self._scores,
            'category': self.category,
            'label': self.label,
            'title': self.title,
            'message': self.message,
            'data': self.data,
            'hide_correctness': self.hide_correctness
        }


"""
Unifying collection of all the commands in ``pedal.assertions``.
"""
from pedal.core.scoring import Score, combine_scores
from pedal.sandbox.commands import call
from pedal.assertions.static import *
from pedal.assertions.runtime import *


def unit_test(function, *tests, else_message=None, score=None, partial_credit=False, **kwargs):
    """
    Helper function for quickly unit testing.

    * Specify exact score for each test case, individually | list[str]
    * Specify exact score for each test case, with single value | str
    * Specify total score for all test cases, no partial credit | False, using `score`
    * Specify total score for all test cases, divide by # of cases passed | True, using `score`

    Args:
        function ():
        *tests ():
        else_message (str): The message to present as a positive if the
             tests all pass.
        score (str): The score to give overall/per test case, depending on
            partial_credit.
        partial_credit (bool or str or list[str]): Whether or not to give credit for
            each unit test as a percentage of the whole (True), or to only give
            points for passing all the tests (False). Defaults to False.
            If a list is passed, those scores will be used per test case.
        **kwargs ():

    Returns:

    """
    if isinstance(partial_credit, bool):
        if partial_credit:
            # Specify total score for all test cases, divide by # of cases passed | True, using `score`
            if score:
                each_score = [str(Score.parse(score) / len(tests)) for test in tests]
            else:
                each_score = [None for test in tests]
        else:
            # Specify total score for all test cases, no partial credit | False, using `score`
            each_score = [None for test in tests]
    elif isinstance(partial_credit, (int, str, float)):
        # Specify exact score for each test case, with single value | str
        each_score = [partial_credit for test in tests]
    else:
        # Specify exact score for each test case, individually | list[str]
        each_score = partial_credit
    with assert_group(function, else_message=else_message, **kwargs) as group_result:
        for test_index, test in enumerate(tests):
            args, expected = test
            assert_equal(call(function, *args), expected, score=each_score[test_index], **kwargs)
    if partial_credit is False:
        group_result.score = score
    else:
        group_result.valence = group_result.POSITIVE_VALENCE
        group_result.score = combine_scores([success.score for success in group_result.successes],
                                            [failure.score for failure in group_result.failures] +
                                            [error.score for error in group_result.errors])
    return not group_result



"""
Simple enumeration of the Modes available for Pedal's command line.
"""
import os
import sys
from contextlib import redirect_stdout
from io import StringIO
import unittest
from textwrap import indent
from unittest.mock import patch
from pprint import pprint

from pedal.command_line.verify import generate_report_out, ReportVerifier
from pedal.core.report import MAIN_REPORT
from pedal.core.submission import Submission
from pedal.utilities.files import normalize_path
from pedal.utilities.text import chomp


def is_progsnap(path):
    """ Determines if this is a ProgSnap file or folder. """
    if os.path.isfile(path):
        name, extension = os.path.splitext(path)
        return extension in ('.zip', '.db')
    else:
        # Does it have all the necessary components?
        return os.path.isfile(os.path.join(path, 'MainTable.csv'))


def get_python_files(paths):
    result = {}
    for path in paths:
        name, ext = os.path.splitext(path)
        if ext != ".py":
            continue
        with open(path, 'r') as submission_file:
            contents = submission_file.read()
        result[path] = contents
    return result


class BundleResult:
    def __init__(self, data, output, error):
        self.data = data
        self.output = output
        self.error = error


class Bundle:
    def __init__(self, script, submission):
        self.script = script
        self.submission = submission
        self.environment = None
        self.result = None

    def run_ics_bundle(self):
        """
        Runs the given instructor control script on the given submission, with the
        accompany contextualizations.
        """
        ics_args = [self.submission, self.environment]
        captured_output = StringIO()
        global_data = {}
        error = None
        if self.environment:
            env = __import__('pedal.environments.'+self.environment,
                             fromlist=[''])
            global_data.update(env.setup_environment(self.submission).fields)
        else:
            MAIN_REPORT.contextualize(self.submission)
        with redirect_stdout(captured_output):
            with patch.object(sys, 'argv', ics_args):
                try:
                    grader_exec = compile(self.script,
                                          self.submission.instructor_file, 'exec')
                    exec(grader_exec, global_data)
                    if ('MAIN_REPORT' in global_data and
                            not global_data['MAIN_REPORT'].resolves and
                            'resolve' in global_data):
                        global_data['resolve']()
                except Exception as e:
                    error = e
        actual_output = captured_output.getvalue()
        self.result = BundleResult(global_data, actual_output, error)


class AbstractPipeline:
    """ Generic pipeline for handling all the phases of executing instructor
    control scripts on submissions, and reformating the output. """
    def __init__(self, config):
        self.config = config
        self.submissions = []
        self.result = None

    def execute(self):
        self.load_submissions()
        self.setup_execution()
        self.run_control_scripts()
        return self.process_output()

    def load_file_submissions(self, scripts):
        # Get instructor control scripts
        all_scripts = []
        for script in scripts:
            script_file_name, script_file_extension = os.path.splitext(script)
            # Single Python file
            if script_file_extension in ('.py',):
                with open(script, 'r') as scripts_file:
                    scripts_contents = scripts_file.read()
                all_scripts.append((script, scripts_contents))
        given_submissions = self.config.submissions
        # If submission is a directory, use it as a directory adjacent to each ics
        if os.path.isdir(given_submissions):
            for script, scripts_contents in all_scripts:
                directory_pattern = given_submissions
                submission_dir = normalize_path(directory_pattern, script)
                submission_files = [
                    os.path.join(submission_dir, sub)
                    for sub in os.listdir(submission_dir)
                ]
                subs = get_python_files(submission_files)
                for main_file, main_code in subs.items():
                    new_submission = Submission(
                        main_file=main_file, main_code=main_code,
                        instructor_file=script
                    )
                    self.submissions.append(Bundle(scripts_contents, new_submission))
        # Otherwise, if the submission is a single file:
        else:
            main_file = given_submissions
            load_error = None
            try:
                with open(given_submissions, 'r') as single_submission_file:
                    main_code = single_submission_file.read()
            except OSError as e:
                # Okay, file does not exist. Load error gets triggered.
                main_code = None
                load_error = e
            for script, scripts_contents in all_scripts:
                new_submission = Submission(
                    main_file=main_file, main_code=main_code,
                    instructor_file=script, load_error=load_error
                )
                self.submissions.append(Bundle(scripts_contents, new_submission))

    def load_progsnap(self, path):
        script_file_name, script_file_extension = os.path.splitext(path)
        if script_file_extension in ('.db',):
            assert "TODO: ProgSnap DB files not yet supported"
            # Progsnap Zip
        elif script_file_extension in ('.zip',):
            assert "TODO: Zip files not yet supported"

    def load_submissions(self):
        given_script = self.config.instructor
        if is_progsnap(given_script):
            self.load_progsnap(given_script)
        elif os.path.isfile(given_script):
            self.load_file_submissions([given_script])
        else:
            python_files = os.listdir(given_script)
            self.load_file_submissions(python_files)


    def setup_execution(self):
        for bundle in self.submissions:
            bundle.environment = self.config.environment

    def run_control_scripts(self):
        for bundle in self.submissions:
            bundle.run_ics_bundle()

    def process_output(self):
        for bundle in self.submissions:
            print(bundle.submission.instructor_file,
                  bundle.submission.main_file,
                  bool(bundle.result.error))


class FeedbackPipeline(AbstractPipeline):
    def process_output(self):
        for bundle in self.submissions:
            print(bundle.submission.instructor_file,
                  bundle.submission.main_file)
            if bundle.result.error:
                print(bundle.result.error)
            else:
                report = bundle.result.data['MAIN_REPORT']
                print(report.result.title)
                print(report.result.message)


class StatsPipeline(AbstractPipeline):
    pass


class VerifyPipeline(AbstractPipeline):
    def process_output(self):
        for bundle in self.submissions:
            bundle.run_ics_bundle()
            if bundle.result.error:
                raise bundle.result.error
            # Now get the expected output
            base = os.path.splitext(bundle.submission.main_file)[0]
            test_name = os.path.basename(base)
            if os.path.exists(base+".out"):
                verifier = ReportVerifier(base+".out", self.config.environment)
                self.add_case(test_name, bundle.result, verifier)
            else:
                if self.config.create_output:
                    report = bundle.result.data['MAIN_REPORT']
                    generate_report_out(base+".out", self.config.environment, report)
                    self.skip_case(test_name)
                else:
                    self.error_case(test_name, bundle.result)
        self.run_cases()

    def __init__(self, config):
        super().__init__(config)

        class TestReferenceSolutions(unittest.TestCase):
            __module__ = 'VerifyTestCase'
            __qualname__ = config.instructor_name
            maxDiff = None
            def __str__(self):
                return f"{config.instructor}, using {self._testMethodName[5:]}.py"
                #return f"{self.id().__name__} (Test{config.instructor})"
            def __repr__(self):
                return str(self)
        self.TestReferenceSolutions = TestReferenceSolutions
        self.tests = []

    def add_case(self, name, bundle_result, verifier):
        """ Add this ``name`` as a new test. """
        if bundle_result.output.strip():
            print(bundle_result.output)
        error = bundle_result.error if bundle_result.error else None
        if error:
            print(error)
        actual = bundle_result.data['MAIN_REPORT'].result.to_json()
        error = bundle_result.error
        expected_fields = verifier.get_final()
        def new_test(self):
            if error:
                raise error
            entire_actual, entire_expected = "", ""
            differing_fields = []
            for field, value in expected_fields:
                self.assertIn(field, actual, msg="Field was not in feedback.")
                actual_value = chomp(str(actual[field]).strip())
                expected_value = chomp(str(value))
                entire_actual += field+": "+actual_value+"\n"
                entire_expected += field+": "+expected_value+"\n"
                if expected_value != actual_value:
                    differing_fields.append(field)
            differing_fields = ', '.join(f"'{field}'" for field in differing_fields)
            self.assertEqual(entire_expected, entire_actual,
                             msg=f"Wrong value for {differing_fields} in {name}.")

        setattr(self.TestReferenceSolutions, "test_" + name, new_test)
        self.tests.append("test_" + name)

    def skip_case(self, name):
        def new_test(self):
            self.skipTest("Output did not exist; created.")
        setattr(self.TestReferenceSolutions, "test_" + name, new_test)
        self.tests.append("test_" + name)

    def error_case(self, name, bundle_result):
        def new_test(self):
            self.fail("Expected output file was not found next to the instructor file. Perhaps you meant to use --create_output?")
        setattr(self.TestReferenceSolutions, "test_" + name, new_test)
        self.tests.append("test_" + name)

    def run_cases(self):
        suite = unittest.defaultTestLoader.loadTestsFromTestCase(self.TestReferenceSolutions)
        runner = unittest.TextTestRunner()
        runner.run(suite)


class GradePipeline(AbstractPipeline):
    def process_output(self):
        for bundle in self.submissions:
            for bundle in self.submissions:
                print(bundle.result.output)
                if bundle.result.error:
                    raise bundle.result.error
                print(bundle.submission.instructor_file,
                      bundle.submission.main_file,
                      bundle.result.data['MAIN_REPORT'].result.score*
                      bundle.result.data['MAIN_REPORT'].result.success,
                      sep=", ")


class SandboxPipeline(AbstractPipeline):
    """ Run the given script in a sandbox. """
    def load_submissions(self):
        self.control_scripts = [
            """
            
            """
        ]

    def process_output(self):
        # Print output
        # Print runtime exceptions, if any
        sandbox = self.result['report']['sandbox']
        output = [sandbox.raw_output]
        if sandbox.error:
            output.append(str(sandbox.error))
        return "\n".join(output)


class DebugPipeline(AbstractPipeline):
    def process_output(self):
        for bundle in self.submissions:
            for bundle in self.submissions:
                print(bundle.submission.instructor_file,
                      bundle.submission.main_file)
                if bundle.result.error:
                    print(bundle.result.error)
                else:
                    print("Output:")
                    print(indent(bundle.result.output, " "*4))
                    report = bundle.result.data['MAIN_REPORT']
                    print("Feedback:")
                    for feedback in report.feedback:
                        print("\t", feedback, repr(feedback.fields))
                    print("Final Feedback")
                    pprint(report.result.to_json())


class MODES:
    """
    The possible modes that Pedal can be run in.
    """
    # Get just the feedback for submissions
    FEEDBACK = 'feedback'
    # Get just the score/correctness
    GRADE = 'grade'
    # Check against expected output
    VERIFY = 'verify'
    # Output all the feedback objects
    STATS = 'stats'
    # Run the file as student code using the sandbox
    SANDBOX = 'sandbox'
    # Output as much information as possible
    DEBUG = 'debug'

    PIPELINES = {
        FEEDBACK: FeedbackPipeline,
        GRADE: GradePipeline,
        STATS: StatsPipeline,
        SANDBOX: SandboxPipeline,
        VERIFY: VerifyPipeline,
        DEBUG: DebugPipeline
    }



"""

.. csv-table:: Cait Report Data
    :header: "Field", "Type", "Initial", "Description"
    :widths: 20, 20, 20, 40

    "ast", "CaitNode", "None", "The CaitNode tree that was most recently parsed out."
    "cache", "{str: CaitNode}", "{}", "A dictionary mapping previously parsed code to CaitNode trees."
    "success", "bool", "True", "Whether the most recent parsing was successful."
    "error", "Exception", "None", "The most recent exception, or None."

"""
from pedal.cait.constants import TOOL_NAME
from pedal.core.commands import system_error
from pedal.core.report import Report, MAIN_REPORT
from pedal.cait.stretchy_tree_matching import StretchyTreeMatcher
from pedal.cait.cait_node import CaitNode
import ast

# TODO: Decide if we want the imports; seems like overkill?
SOURCE_TOOL_NAME = 'source'
TIFA_TOOL_NAME = 'tifa'


class CaitException(Exception):
    pass


def _parse_source(code, report=MAIN_REPORT):
    """
    Parses the given code and returns its Cait representation. If the parse was
    unsuccessful, it attaches the error to the report.

    Args:
        code (str): A string of Python code.
        report (Report): A Report to store information in.
    Returns:
        AstNode: The parsed AST reprensetation, or None
    """
    try:
        parsed = ast.parse(code)
    except SyntaxError as e:
        system_error(TOOL_NAME, "Could not parse code:" + str(e), report=report)
        report[TOOL_NAME]['success'] = False
        report[TOOL_NAME]['error'] = e
        return ast.parse("")
    return parsed


def reparse_if_needed(student_code=None, report=MAIN_REPORT):
    """
    Retrieves the current report for CAIT. If there is no CAIT report, it will
    generate one. If source code is given, that will be used instead of the
    report's submission.

    Args:
        student_code (str): The code to parse into the a CaitNode tree. If
            None, then it will use the code in the report's submission.
        report (Report): The report to attach data to.

    Returns:
        dict: Returns the Cait Report
    """
    cait = report[TOOL_NAME]
    if student_code is not None:
        if student_code in cait['cache']:
            cait['ast'] = cait['cache'][student_code]
            return cait
        else:
            student_ast = _parse_source(student_code, report=report)
    else:
        student_code = report.submission.main_code
        # Have we already parsed this code?
        if student_code in cait['cache']:
            cait['ast'] = cait['cache'][student_code]
            return cait
        # Try to steal parse from Source module, if available
        if report[SOURCE_TOOL_NAME]['success']:
            student_ast = report[SOURCE_TOOL_NAME]['ast']
        else:
            student_ast = _parse_source(student_code, report=report)
    cait['ast'] = cait['cache'][student_code] = CaitNode(student_ast, report=report)
    return cait


def require_tifa(self):
    """
    Confirms that TIFA was run successfully, otherwise raises a
    CaitException.

    TODO: Is this deprecated?
    """
    if not self.report[TIFA_TOOL_NAME]['success']:
        system_error(TOOL_NAME, "TIFA was not successfully run prior to CAIT.")
        raise CaitException("TIFA was not run prior to CAIT.")


# noinspection PyBroadException
def parse_program(student_code=None, report=MAIN_REPORT):
    """
    Parses student code and produces a CAIT representation.

    Args:
        student_code (str): The student source code to parse. If None, defaults
            to the code within the Source tool of the given Report.
        report (Report): The report to attach data to. Defaults to MAIN_REPORT.

    Returns:
        CaitNode: A CAIT-enhanced representation of the root Node.
    """
    cait_report = reparse_if_needed(student_code, report=report)
    return cait_report['ast']


def expire_cait_cache(report=MAIN_REPORT):
    """
    Deletes the most recent CAIT run and any cached CAIT parses.

    Args:
        report (Report): The report to attach data to. Defaults to MAIN_REPORT.
    """
    report['cait']['ast'] = None
    report['cait']['cache'] = {}


def def_use_error(node, report=MAIN_REPORT):
    """
    Checks if node is a name and has a def_use_error

    Args:
        node (str or AstNode or CaitNode): The Name node to look up.
        report (Report): The report to attach data to. Defaults to MAIN_REPORT.
    Returns:
        True if the given name has a def_use_error
    """
    if not isinstance(node, str) and node.ast_name != "Name":
        raise TypeError
    from pedal.tifa.commands import get_issues
    from pedal.tifa.feedbacks import initialization_problem
    def_use_issues = get_issues(initialization_problem)
    if not isinstance(node, str):
        node_id = node.id
    else:
        node_id = node
    has_error = False
    for issue in def_use_issues:
        name = issue.fields['name']
        if name == node_id:
            has_error = True
            break
    return has_error


# noinspection PyBroadException
def data_state(node, report=MAIN_REPORT):
    """
    Determines the Tifa State of the given node.

    Args:
        node (str or AstNode or CaitNode): The Name node to look up in TIFA.
        report (Report): The report to attach data to. Defaults to MAIN_REPORT.
    Returns:
        The State of the object (Tifa State) or None if it doesn't exist
    """
    if not isinstance(node, str) and node.ast_name != "Name":
        raise TypeError
    if isinstance(node, str):
        node_id = node
    else:
        node_id = node.id
    try:
        return report[TIFA_TOOL_NAME]["latest"].top_level_variables[node_id]
    except KeyError:
        return None


def data_type(node, report=MAIN_REPORT):
    """
    Looks up the type of the node using Tifa's analysis.

    Args:
        node (str or AstNode or CaitNode): The Name node to look up in TIFA.
        report (Report): The report to attach data to. Defaults to MAIN_REPORT.
    Returns:
        The type of the object (Tifa type) or None if a type doesn't exist
    """
    state = data_state(node, report=report)
    if state is not None:
        return state.type
    return None


def find_match(pattern, student_code=None, report=MAIN_REPORT, cut=False, use_previous=None):
    """
    Apply Tree Inclusion and return the first match of the `pattern` in the
    `student_code`.

    Args:
        pattern (str): The CaitExpression to match against.
        student_code (str): The string of student code to check against.
            Defaults to the code of the Source tool in the Report.
        report (Report): The report to attach data to.
        cut (bool): Set to true to trim root to first branch
        use_previous (AstMap): If user wants to continue off of a previously found match
    Returns:
        CaitNode or None: The first matching node for the given pattern, or
            None if nothing was found.
    """
    matches = find_matches(pattern=pattern, student_code=student_code,
                           report=report, cut=cut, use_previous=use_previous)
    if matches:
        return matches[0]
    else:
        return None


def find_matches(pattern, student_code=None, cut=False, report=MAIN_REPORT, use_previous=None):
    """
    Apply Tree Inclusion and return all matches of the `pattern` in the
    `student_code`.

    Args:
        pattern (str): The CaitExpression to match against.
        student_code (str): The string of student code to check against.
        report (Report): The report to attach data to.
        cut (bool): Set to true to trim root to first branch
        use_previous (AstMap): If user wants to continue off of a previously found match
    Returns:
        list[pedal.cait.ast_map.AstMap]: All matching nodes for the given pattern.
    """
    cait_report = reparse_if_needed(student_code, report)
    if not cait_report['success']:
        return []
    student_ast = cait_report['ast']
    matcher = StretchyTreeMatcher(pattern, report=report)
    return matcher.find_matches(student_ast, use_previous=use_previous)


def find_submatches(pattern, student_code, is_mod=False, report=MAIN_REPORT):
    """
    Incomplete.
    """
    return find_expr_sub_matches(pattern, student_code, is_mod, report=report)


def find_expr_sub_matches(pattern, student_code, is_mod=False, report=MAIN_REPORT):
    """
    Finds pattern in student_code
    # TODO: Add code to make pattern accept CaitNodes
    # TODO: Make this function without so much meta knowledge
    Args:
        pattern: the expression to find (str that MUST evaluate to a Module node with a single child or an AstNode)
        student_code: student subtree
        is_mod (bool): currently hack for multiline sub matches
        report: defaults to MAIN_REPORT unless another one exists
    Returns:
        a list of matches or False if no matches found
    """
    is_node = isinstance(pattern, CaitNode)
    if not isinstance(pattern, str) and not is_node:
        raise TypeError("pattern expected str or CaitNode, found {0}".format(type(pattern)))
    matcher = StretchyTreeMatcher(pattern, report=report)
    if (not is_node and not is_mod) and len(matcher.root_node.children) != 1:
        raise ValueError("pattern does not evaluate to a singular statement")
    return matcher.find_matches(student_code, check_meta=False)


def find_asts(node_name: str, student_code=None, report=MAIN_REPORT):
    """
    Find all occurrences of the given AST node, based on the name of the AST
    node (e.g., `"For"` or `"FunctionDef"`).

    Args:
        node_type: the string representing the "type" of node to look for
        student_code (str): Optionally, different code to parse and search.
        report: defaults to MAIN_REPORT unless another one is given.

    Returns:
        a list of Ast Nodes (cait_nodes) of self that are of the specified type (including self if self
                meets that criteria)

    Returns:

    """
    cait_report = reparse_if_needed(student_code, report)
    if not cait_report['success']:
        return []
    student_ast = cait_report['ast']
    return student_ast.find_all(node_name)


def reset(report=MAIN_REPORT):
    """

    Args:
        report:

    Returns:

    """
    report[TOOL_NAME] = {
        'success': True,
        'error': None,
        'ast': None,
        'cache': {}
    }
    return report[TOOL_NAME]


Report.register_tool(TOOL_NAME, reset)



"""
Resolvers are called at the end of the grading script in order to determine
what pieces of feedback should actually be delivered.
"""

__all__ = ['Resolver']

from pedal.core.report import MAIN_REPORT


class Resolver:
    """
    A resolver is responsible for filtering, ordering, and choosing Feedback.

    Ideally, it is not responsible for transforming - that is a job for the Environment.
    """

    def resolve(self):
        """
        Actually choose the feedback that will be returned
        """

    def _on_pre_resolve(self):
        """
        Hook event that is run before a resolver actually begins its process.

        TODO: This doesn't do anything yet
        """

    def _on_post_resolve(self):
        """
        Hook event that is after a resolver actually begins its process.
        """



"""

.. csv-table:: Source Report Data
    :header: "Field", "Type", "Initial", "Description"
    :widths: 20, 20, 20, 40

    "substitutions", "List[Substitution]", "[]", "The history of previous source codes before the latest substitutions."
    "success", "bool", "None", "Whether the current file has been parsed successfully."
    "ast", "ast.Ast", "None", "The root node of the latest successfully parsed chunk of code."

"""

import sys
import ast

from pedal.core.feedback import CompositeFeedbackFunction
from pedal.core.report import Report, MAIN_REPORT
from pedal.core.submission import Submission
from pedal.source.constants import TOOL_NAME, DEFAULT_STUDENT_FILENAME
from pedal.source.sections import separate_into_sections
from pedal.source.feedbacks import blank_source, syntax_error, source_file_not_found
from pedal.source.substitutions import Substitution


def reset(report=MAIN_REPORT):
    """
    Resets the Source tool back to its initial configuration, removing any
    previously stored code/parses/section info.

    Args:
        report: The report object to be reseting.

    Returns:
        dict: The data associated with the Source tool.
    """
    report[TOOL_NAME] = {
        'substitutions': [],

        'success': None,
        'ast': None,

        'independent': None,
        'sections': None,
        'section': None,
        'section_pattern': None,
    }
    return report[TOOL_NAME]


Report.register_tool(TOOL_NAME, reset)


def set_source(code, filename=DEFAULT_STUDENT_FILENAME, sections=False,
               independent=False, report=MAIN_REPORT):
    """
    Sets the contents of the Source to be the given code. Can also be
    optionally given a filename.

    Args:
        code (str): The contents of the source file.
        filename (str): The filename of the students' code. Defaults to
                        `answer.py`.
        sections (str or bool): Whether or not the file should be divided into
                                sections. If a str, then it should be a
                                Python regular expression for how the sections
                                are separated. If False, there will be no
                                sections. If True, then the default pattern
                                will be used: '^##### Part (\\d+)$'
        independent (bool): Whether the separate sections should be considered
            separate or all existing in an accumulating namespace.
        report (Report): The report object to store data and feedback in. If
                         left None, defaults to the global MAIN_REPORT.
    """
    if report.submission is None:
        report.contextualize(Submission({filename: code}, filename, code))
    else:
        backup = Substitution(report.submission.main_code, report.submission.main_file)
        report[TOOL_NAME]['substitutions'].append(backup)
        report.submission.replace_main(code, filename)

    report[TOOL_NAME]['independent'] = independent
    report[TOOL_NAME]['success'] = True
    if not sections:
        report[TOOL_NAME]['sections'] = None
        report[TOOL_NAME]['section'] = None
        verify(code, report=report)
    else:
        separate_into_sections(report=report)


# TODO: source_prepend and source_append


def restore_code(report=MAIN_REPORT):
    """

    Args:
        report:
    """
    if TOOL_NAME in report:
        old_submission = report[TOOL_NAME]['substitutions'].pop()
        report.submission.replace_main(old_submission.code, old_submission.filename)
        verify(report=report)


@CompositeFeedbackFunction(blank_source, syntax_error)
def verify(code=None, filename=DEFAULT_STUDENT_FILENAME, report=MAIN_REPORT,
           muted=False):
    """
    Parses the given source code and checks for syntax errors; if no code is given, defaults to the
    current Main file of the submission.

    Args:
        muted:
        code (str): Some code to parse and syntax check.
        filename: An optional filename to use
        report:

    Returns:

    """
    if code is None:
        code = report.submission.main_code
        filename = report.submission.main_file
    if report.submission.load_error:
        source_file_not_found(filename, None)
        report[TOOL_NAME]['success'] = False
        return False
    if code.strip() == '':
        blank_source()
        report[TOOL_NAME]['success'] = False
    try:
        parsed = ast.parse(code, filename)
        report[TOOL_NAME]['ast'] = parsed
    except SyntaxError as e:
        syntax_error(e.lineno, e.filename, code, e.offset, e,
                     sys.exc_info(), report=report, muted=muted)
        report[TOOL_NAME]['success'] = False
        report[TOOL_NAME]['ast'] = ast.parse("")
    report[TOOL_NAME]['success'] = True
    return report[TOOL_NAME]['success']


# Legacy verify_section; now done by verify since its aware of sections
verify_section = verify


def get_program(report=MAIN_REPORT) -> str:
    """
    Retrieves the current main file's code.
    """
    return report.submission.main_code


def get_original_program(report=MAIN_REPORT) -> str:
    """
    Retrieves the original version of the Submission's main code, ignoring all subsitutions.
    Args:
        report:

    Returns:

    """
    return report[TOOL_NAME]['substitutions'][0].code


@CompositeFeedbackFunction(source_file_not_found)
def set_source_file(filename: str, sections=False, independent=False, report=MAIN_REPORT):
    """
    Uses the given `filename` on the filesystem as the new main file.

    Args:
        filename:
        sections:
        independent:
        report:

    Returns:

    """
    try:
        with open(filename, 'r') as student_file:
            set_source(student_file.read(), filename=filename,
                       sections=sections, independent=independent,
                       report=report)
    except IOError:
        source_file_not_found(filename, sections, report=report)
        report[TOOL_NAME]['success'] = False



"""
Feedback functions of the Source module
"""

from pedal.core.commands import feedback
from pedal.core.feedback import FeedbackResponse
from pedal.core.location import Location
from pedal.core.report import MAIN_REPORT
from pedal.utilities.exceptions import ExpandedTraceback
from pedal.source.constants import TOOL_NAME


class SourceFeedback(FeedbackResponse):
    """ Base class of all Feedback functions for Source Tool """
    category = feedback.CATEGORIES.SYNTAX
    kind = feedback.KINDS.MISTAKE
    tool = TOOL_NAME


class blank_source(SourceFeedback):
    """ Source code file was blank. """
    title = "No Source Code"
    message_template = "Source code file is blank."
    justification = "After stripping the code, there were no characters."


class not_enough_sections(SourceFeedback):
    """ Didn't have all the needed sections. """
    title = "Not Enough Sections"
    message_template = ("Tried to advance to next section but the "
                        "section was not found. Tried to load section "
                        "{count}, but there were only {found} sections.")
    justification = ("Section index exceeded the length of the separated "
                     "sections list.")

    def __init__(self, section_number, found, **kwargs):
        fields = {'count': section_number, 'found': found}
        super().__init__(fields=fields, **kwargs)


class source_file_not_found(SourceFeedback):
    """ No source file was given. """
    title = 'Source File Not Found'
    message_template = ("The given filename {name:filename} was either not "
                        "found or could not be opened. Please make sure the "
                        "file is available.")
    version = '0.0.1'
    justification = "IOError while opening file to set_source"

    def __init__(self, name, sections, **kwargs):
        report = kwargs.get("report", MAIN_REPORT)
        fields = {'name': name, 'sections': sections }
        group = 0 if sections else kwargs.get('group')
        super().__init__(fields=fields, group=group, **kwargs)


class syntax_error(SourceFeedback):
    """ Generic feedback for any kind of syntax error. """
    muted = False
    title = "Syntax Error"
    message_template = ("Bad syntax on line {lineno:line}\n\n"
                        "The traceback was:\n{traceback_message}\n\n"
                        "Suggestion: Check line {lineno:line}, the line "
                        "before it, and the line after it.")
    version = '0.0.1'
    justification = "Syntax error was triggered while calling ast.parse"

    def __init__(self, line, filename, code, col_offset,
                 exception, exc_info, **kwargs):
        report = kwargs.get('report', MAIN_REPORT)
        files = report.submission.get_files_lines()
        if filename not in files:
            files[filename] = code.split("\n")
        if report.submission is not None:
            lines = report.submission.get_lines()
            line_offsets = report.submission.line_offsets
        else:
            lines = code.split("\n")
            line_offsets = {}
        line_offset = line_offsets.get(filename, 0)
        traceback = ExpandedTraceback(exception, exc_info, False,
                                      [report.submission.instructor_file],
                                      line_offsets, [filename], lines, files)
        traceback_stack = traceback.build_traceback()
        traceback_message = traceback.format_traceback(traceback_stack,
                                                       report.format)
        line_offset = line_offsets.get(filename, 0)
        fields = {'lineno': line+line_offset,
                  'filename': filename, 'offset': col_offset,
                  'exception': exception,
                  'traceback': traceback,
                  'traceback_stack': traceback_stack,
                  'traceback_message': traceback_message}
        location = Location(line=line+line_offset, col=col_offset, filename=filename)
        super().__init__(fields=fields, location=location, **kwargs)


class incorrect_number_of_sections(SourceFeedback):
    """ Incorrect number of sections """
    title = "Incorrect Number of Sections"
    message_template = ("Incorrect number of sections in your file. "
                     "Expected {count}, but only found {found}")
    justification = ""

    def __init__(self, count, found, **kwargs):
        fields = {'count': count, 'found': found}
        super().__init__(fields=fields, **kwargs)

# TODO: IndentationError
# TODO: TabError



import string
import re

# Number encapsulates bool, int, float, complex, decimal.Decimal, etc.
try:
    from numbers import Number
except:
    Number = (bool, int, float, complex)

try:
    bytes
except NameError:
    bytes = str

try:
    frozenset()
except:
    frozenset = tuple()

try:
    punctuation_table = str.maketrans(string.punctuation, ' ' * len(string.punctuation))
except AttributeError:
    punctuation_table = None

if punctuation_table is None:
    def strip_punctuation(a_string):
        """

        Args:
            a_string:

        Returns:

        """
        return ''.join(ch for ch in a_string if ch not in set(string.punctuation))
else:
    def strip_punctuation(a_string):
        """

        Args:
            a_string:

        Returns:

        """
        return a_string.translate(punctuation_table)

SET_GENERATOR_TYPES = (type({}.keys()), type({}.values()), type({}.items()))

LIST_GENERATOR_TYPES = (type(map(bool, [])), type(filter(bool, [])),
                        type(range(0)), type(reversed([])), type(zip()),
                        type(enumerate([])))


def _split_into_lolos(a_string) -> 'List[List[str]]':
    """
    Splits the given string into a list of list of strings, based on new lines.

    Args:
        a_string: The string to split

    Returns: The list of list of strings
    """
    return [list(line) for line in a_string.split("\n")]


def _normalize_string(a_string, numeric_endings=False):
    # Lower case
    a_string = a_string.lower()
    # Remove trailing decimals (TODO: How awful!)
    if numeric_endings:
        a_string = re.sub(r"(\s*[0-9]+)\.[0-9]+(\s*)", r"\1\2", a_string)
    # Remove punctuation
    a_string = strip_punctuation(a_string)
    # Split lines
    lines = a_string.split("\n")
    normalized = [[piece for piece in line.split()]
                  for line in lines]
    normalized = [[piece for piece in line if piece]
                  for line in normalized if line]
    return sorted(normalized)


def output_test(actual, expected, _exact_strings):
    """

    Args:
        actual:
        expected:
        _exact_strings: Force capitalization/whitespace/symbols to match exactly.

    Returns:

    """
    normalizer = _normalize_string if not _exact_strings else _split_into_lolos
    normalized_actual = [normalizer(line) for line in actual]
    if isinstance(expected, (str, bytes)):
        return normalizer(expected) in normalized_actual
    else:
        normalized_expected = [normalizer(line) for line in expected]
        return all(each_actual in normalized_expected for each_actual in normalized_actual)


def equality_test(actual, expected, _exact_strings, _delta):
    """

    Args:
        actual:
        expected:
        _exact_strings:
        _delta:

    Returns:

    """
    # Check if generators
    if isinstance(expected, LIST_GENERATOR_TYPES):
        expected = list(expected)
    elif isinstance(expected, SET_GENERATOR_TYPES):
        expected = set(expected)
    if isinstance(actual, LIST_GENERATOR_TYPES):
        actual = list(actual)
    elif isinstance(actual, SET_GENERATOR_TYPES):
        actual = set(actual)

    # Float comparison
    if isinstance(expected, float) and isinstance(actual, (float, int)):
        error = 10 ** (-_delta)
        return abs(expected - actual) < error
    # Other numerics
    elif isinstance(expected, Number) and isinstance(actual, Number) and isinstance(expected, type(actual)):
        return expected == actual
    # String comparisons
    elif ((isinstance(expected, str) and isinstance(actual, str)) or
          (isinstance(expected, bytes) and isinstance(actual, bytes))):
        if _exact_strings:
            return expected == actual
        else:
            return _normalize_string(expected) == _normalize_string(actual)
    # Exact Comparison
    elif actual == expected:
        return True
    # Sequence comparisons
    elif isinstance(expected, list) and isinstance(actual, list):
        return _are_sequences_equal(actual, expected, _exact_strings, _delta)
    elif isinstance(expected, tuple) and isinstance(actual, tuple):
        return _are_sequences_equal(actual, expected, _exact_strings, _delta)
    elif isinstance(expected, set) and isinstance(actual, set):
        return _are_sets_equal(actual, expected, _exact_strings, _delta)
    elif isinstance(expected, frozenset) and isinstance(actual, frozenset):
        return _are_sets_equal(actual, expected, _exact_strings, _delta)
    elif isinstance(expected, dict) and isinstance(actual, dict):
        primary_keys = set(expected.keys())
        if not _are_sets_equal(primary_keys, set(actual.keys()), _exact_strings, _delta):
            return False
        for key in primary_keys:
            if not equality_test(expected[key], actual[key], _exact_strings, _delta):
                return False
        return True
    # Else
    return False


def _are_sequences_equal(x, y, _exact_strings, _delta):
    """
    For sequences that support __len__, __iter__, and should have the same
    order.
    """
    if len(x) != len(y):
        return False
    for x_element, y_element in zip(x, y):
        if not equality_test(x_element, y_element, _exact_strings, _delta):
            return False
    return True


def _set_contains(needle, haystack, _exact_strings, _delta):
    """
    Tests if the given needle is one of the elements of haystack, using
    the _is_equal function.
    """
    for element in haystack:
        if equality_test(element, needle, _exact_strings, _delta):
            return True
    return False


def _are_sets_equal(x, y, _exact_strings, _delta):
    """
    For sequences that support __len__, __iter__, but order does not matter.
    """
    if len(x) != len(y):
        return False
    for x_element in x:
        if not _set_contains(x_element, y, _exact_strings, _delta):
            return False
    return True


def iterable(obj) -> bool:
    """
    Determines if this object satisfies the iterable interface, which
    requires it to support either ``__iter__`` or ``__getitem__``.
    """
    return hasattr(obj, '__iter__') or hasattr(obj, '__getitem__')



"""
Environment to support BlockPy.

# Support our sysmodules hack by clearing out any lingering old data
from pedal.core.report import MAIN_REPORT
MAIN_REPORT.clear()

# Interface for interacting with BlockPy
from utility import *

from pedal.environments.blockpy import setup_pedal
pedal = setup_pedal(skip_tifa=${skip_tifa},
                    main_file='answer.py',
                    main_code=${safeCode})

# Execute students' code
if not get_model_info('assignment.settings.disableInstructorRun'):
    set_inputs(get_model_info('execution.input'))
    student = run()
else:
    student = get_sandbox()

# Load in some commonly used tools
from pedal.cait.cait_api import parse_program
from pedal.sandbox.commands import *

# TODO: Refactor resolver to return instructions
# Monkey-patch questions
from pedal import questions
questions.show_question = set_instructions

# Run the actual instructor code
${instructorCode}

# Resolve everything
final = pedal.resolve()
SUCCESS = final.success
SCORE = final.score
CATEGORY = final.category
LABEL = final.title
MESSAGE = final.message
DATA = final.data
HIDE = final.hide_correctness
"""
from pedal import verify
from pedal.core.environment import Environment
from pedal.core.report import MAIN_REPORT
from pedal.sandbox import run, get_sandbox, set_input, start_trace
from pedal.tifa import tifa_analysis
from pedal.resolvers.simple import resolve


class BlockPyEnvironment(Environment):
    """
    Configures the BlockPy programming environment.
    """
    def __init__(self, files=None, main_file='answer.py', main_code=None,
                 user=None, assignment=None, course=None, execution=None,
                 instructor_file='on_run.py', skip_tifa=False, skip_run=False,
                 inputs=None, set_success=True,
                 report=MAIN_REPORT, trace=True):
        super().__init__(files=files, main_file=main_file, main_code=main_code,
                         user=user, assignment=assignment, course=course,
                         execution=execution, instructor_file=instructor_file,
                         report=report)
        verify(report=self.report)
        if not skip_tifa:
            tifa_analysis(report=self.report)
        if inputs:
            set_input(inputs)
        if skip_run:
            student = get_sandbox(report=report)
        else:
            if trace:
                start_trace()
            student = run(report=report)
        self.fields = {
            'student': student,
            'resolve': resolve
        }


setup_environment = BlockPyEnvironment

"""
TODO: Enhance the following

Override formatters
assertions could get HTML diffs
Grouping a bunch of assert_equal produces a table
"""



"""
A simple environment for quickly running Pedal code.
"""

import sys

from pedal.assertions import resolve_all
from pedal.core.environment import Environment
from pedal.core.report import MAIN_REPORT
from pedal.source import verify
from pedal.cait import parse_program
from pedal.sandbox.commands import run
from pedal.resolvers import simple


def parse_argv():
    """ Retrieve fields from sys.argv """
    if len(sys.argv) == 2:
        main_file = sys.argv[1]
        with open(main_file) as main_file_handle:
            main_code = main_file_handle.read()
        return sys.argv[0], {main_file: main_code}, main_file, main_code, None, None, None, None
    # elif len(sys.argv) == 3:
    else:
        return sys.argv
    # TODO: Finish handling arguments intelligently


class StandardEnvironment(Environment):
    """
    The StandardEnvironment is a convenience class for basic command line
    usage of Pedal. It doesn't do very much, but runs a few of our tools:
    CAIT, TIFA, and Sandbox. It uses the Simple resolver, and just dumps the
    result to the command line.

    from pedal.environments.standard import setup_pedal
    code, student, resolve = setup_pedal()
    """
    def __init__(self, files=None, main_file='answer.py', main_code=None,
                 user=None, assignment=None, course=None, execution=None,
                 instructor_file='on_run.py', skip_tifa=False, set_success=True,
                 report=MAIN_REPORT):
        # Possibly user passed in stuff via the command line.
        if files is None and main_code is None:
            (instructor_file, files, main_file, main_code, user, assignment,
             course, execution) = parse_argv()
        super().__init__(files=files, main_file=main_file, main_code=main_code,
                         user=user, assignment=assignment, course=course,
                         execution=execution, instructor_file=instructor_file,
                         report=report)
        # Then default custom stuff
        verify(report=report)
        self.ast = parse_program(report=report)
        if skip_tifa:
            self.tifa = None
        else:
            from pedal.tifa import tifa_analysis
            self.tifa = tifa_analysis(report=report)
        self.student = run(threaded=True, report=report)
        self.set_success = set_success

    def print_resolve(self, *args, **kwargs):
        """
        Trivial formatter for resolver, just dumps the
        Title/Label/Score/Message. Any arguments are forwarded to
        :py:func:`pedal.resolvers.simple.resolve`
        """
        #resolve_all(set_successful=self.set_success, no_phases_is_success=True,
        #            report=self.report)
        # TODO: Fix resolve_all
        result = simple.resolve(*args, **kwargs)
        # print("Feedback Label:", result.label)
        print("Title:", result.title)
        print("Label:", result.label)
        print("Score:", result.score)
        print("Message:", result.message)
        return result

    def get_fields(self):
        """ Returns Cait's AST, the student Sandbox,
            and the Simple Resolver. """
        return self.ast, self.student, self.print_resolve


setup_pedal = StandardEnvironment



"""
The collection of top-level commands for the Sandbox module. Note that
most of these simply act on the current MAIN_REPORT's Sandbox instance, without
doing any logic themselves.
"""

from pedal.core.report import MAIN_REPORT
from pedal.sandbox.constants import TOOL_NAME
from pedal.sandbox.sandbox import Sandbox
from pedal.source.constants import TOOL_NAME as SOURCE_TOOL_NAME
from pedal.utilities.ast_tools import FindExecutableLines


def run(code=None, filename=None, inputs=None, threaded=None,
        after=None, before=None, report=MAIN_REPORT):
    """
    If both ``code`` and ``filename`` are None, then the submission's
    main file will be executed. If ``code`` is given but ``filename`` is
    not, then it is assumed to be instructor code.

    Args:
        code (str or :py:class:`~pedal.cait.cait_node.CaitNode` or None):
            The code to execute.
        filename (str or None): The filename to use for this code.
        inputs (list[str]): Optional inputs to be passed to
            :py:func:`~pedal.sandbox.Sandbox.set_input`.
        threaded (bool): Whether or not to run this code in a threaded
            environment, which allows for timeouts.
        after (str): Code to run after this code (without a filename).
        before (str): Code to run before this code (without a filename).
        report (:py:class:`pedal.core.report.Report`): The report with the
            sandbox instance.

    Returns:
        Sandbox: The sandbox instance that this was run in.
    """
    sandbox: Sandbox = report[TOOL_NAME]['sandbox']
    return sandbox.run(code=code, filename=filename, inputs=inputs,
                       threaded=threaded, after=after, before=before)


def call(function, *args, target="_", threaded=None, inputs=None,
         function_kwargs=None, args_locals=None, kwargs_locals=None,
         report=MAIN_REPORT, **kwargs):
    """
    Execute the given function from the student's namespace.

    The ``args_locals`` and ``kwargs_locals`` values allow you to
    use student's local variables as arguments, instead of literal values.
    They actually support any arbitrary Python code, it will be injected
    without modification.

    Args:
        function (str): The name of the function to call.
        *args (Any): Any number of positional arguments to be passed to
            the called function.
        target (str): The variable to assign the result of this call to.
            Defaults to ``"_"``.
        threaded (bool): Whether or not to run this code in a threaded
            environment, which allows for timeouts.
        inputs (str or list[str]):
        function_kwargs (dict[str, Any]): Additional keyword arguments
            that could not be passed in directly via ``**kwargs`` (perhaps
            because they conflict with an existing parameter like
            ``target``).
        args_locals (list[str]): A list of names (or any valid Python
            expression) that will be passed as positional arguments to
            the function (as opposed to actual values). If any value is
            None (or if the list is too short), then the corresponding
            position argument from ``*args`` will be used instead.
        kwargs_locals (dict[str, str]): A dictionary of keyword argument
            names mapped to local names (or any valid Python expression),
            that will be used as keyword parameters similar to
            ``args_locals``.
        **kwargs (): Additional keyword arguments passed to the called
            function.
        report (:py:class:`pedal.core.report.Report`): The report with the
            sandbox instance.

    Returns:
        Exception or :py:class:`~pedal.sandbox.sandbox.SandboxResult`: The
            result of calling the function will be returned, proxied behind
            a SandboxResult (which attempts to perfectly emulate that
            value). If the function call failed, the exception will be
            returned instead.
    """
    sandbox: Sandbox = report[TOOL_NAME]['sandbox']
    return sandbox.call(function, *args, target=target, threaded=threaded,
                        inputs=inputs, function_kwargs=function_kwargs,
                        args_locals=args_locals, kwargs_locals=kwargs_locals,
                        **kwargs)


def evaluate(code, target="_", threaded=None, report=MAIN_REPORT):
    """
    Evaluates the given code and assigns the result to the given target.
    Will cause an error if ``code`` is not a valid expression.

    Args:
        code (str or :py:class:`~pedal.cait.cait_node.CaitNode`):
            The code to execute. If a CaitNode, then that will be executed
            directly instead of being compiled.
        target (str): The name of the variable to assign the result to.
            Note that the result is also returned by this function.
        threaded (bool): Whether or not to run this code in a threaded
            environment, which allows for timeouts.
        report (:py:class:`pedal.core.report.Report`): The report with the
            sandbox instance.

    Returns:
        Exception or :py:class:`~pedal.sandbox.sandbox.SandboxResult`: The
            result of evaluating the code will be returned, proxied behind
            a SandboxResult (which attempts to perfectly emulate that
            value). If the function call failed, the exception will be
            returned instead.
    """
    sandbox: Sandbox = report[TOOL_NAME]['sandbox']
    return sandbox.evaluate(code, target=target, threaded=threaded)


def clear_input(report=MAIN_REPORT):
    """ Removes any existing inputs set up for the sandbox. """
    sandbox: Sandbox = report[TOOL_NAME]['sandbox']
    sandbox.clear_input()


def queue_input(*inputs, report=MAIN_REPORT):
    """
    Adds the given ``*inputs`` to be used as input during subsequent
        executions (e.g., :py:func:`pedal.sandbox.commands.run`) if the
        student's code calls :py:func:`input`. This does not remove existing
        inputs.

    Args:
        *inputs (str): One or more strings that will be set. The first string
            passed in is the first string that will be passed in as input.
        report (:py:class:`pedal.core.report.Report`): The report with the
            sandbox instance.
    """
    sandbox: Sandbox = report[TOOL_NAME]['sandbox']
    sandbox.set_input(inputs, clear=False)


def set_input(inputs, clear=True, report=MAIN_REPORT):
    """
    Sets the given ``inputs`` to be used as input during subsequent
        executions (e.g., :py:func:`pedal.sandbox.commands.run`) if the
        student's code calls :py:func:`input`. Unless the ``clear`` parameter
        is set to False, this removes existing inputs.

    Args:
        inputs (str or list[str]): One or more strings that will be set. The
            first string passed in is the first string that will be passed in
            as input.
        clear (bool): Whether or not to remove any existing inputs set up
            in the sandbox.
        report (:py:class:`pedal.core.report.Report`): The report with the
            sandbox instance.
    """
    sandbox: Sandbox = report[TOOL_NAME]['sandbox']
    sandbox.set_input(inputs, clear=clear)


def get_input(report=MAIN_REPORT):
    """
    Retrieves the current inputs that are available for execution.

    Args:
        report (:py:class:`pedal.core.report.Report`): The report with the
            sandbox instance.

    Returns:
        list[str]: The current list of inputs available for execution. The first
            element of this list is the next string that will be passed to the
            next student call of :py:func:`input`.
    """
    sandbox: Sandbox = report[TOOL_NAME]['sandbox']
    return sandbox.inputs


def clear_output(report=MAIN_REPORT):
    """ Removes any existing outputs associated with the sandbox. """
    sandbox: Sandbox = report[TOOL_NAME]['sandbox']
    sandbox.clear_output()


def get_output(report=MAIN_REPORT):
    """ Retrieves the current output (whatever the student has printed) since
    execution began. """
    sandbox: Sandbox = report[TOOL_NAME]['sandbox']
    return sandbox.output


def get_exception(report=MAIN_REPORT):
    """ Returns an exception if one occurred during the last execution,
    otherwise returns None. """
    sandbox: Sandbox = report[TOOL_NAME]['sandbox']
    return sandbox.exception


def clear_student_data(report=MAIN_REPORT):
    """ Removes any data in the student namespace. """
    sandbox: Sandbox = report[TOOL_NAME]['sandbox']
    sandbox.clear_data()


def get_student_data(report=MAIN_REPORT):
    """
    Retrieves the current data in the student namespace. Note that this is
    the data itself - modifying the dictionary will modify the data in the
    students' namespace for subsequent executions!

    Args:
        report (:py:class:`pedal.core.report.Report`): The report with the
            sandbox instance.

    Returns:
        dict[str, Any]: The student's data namespace, mapping names to the
            values themselves.
    """
    sandbox: Sandbox = report[TOOL_NAME]['sandbox']
    return sandbox.data


def get_sandbox(report=MAIN_REPORT):
    """
    Retrieves the current sandbox instance attached to this report.
    Typically, this is used to retrieve the sandbox without running the
    students' code.

    Args:
        report (:py:class:`pedal.core.report.Report`): The report with the
            sandbox instance.

    Returns:
        :py:class:`pedal.sandbox.Sandbox`: The sandbox instance.
    """
    return report[TOOL_NAME]['sandbox']


def clear_sandbox(report=MAIN_REPORT):
    """ Removes any existing data within the current sandbox instance. """
    sandbox: Sandbox = report[TOOL_NAME]['sandbox']
    sandbox.clear()


def get_trace(report=MAIN_REPORT):
    """
    Retrieves the set of line that have been traced (recognized as executed).

    Args:
        report (:py:class:`pedal.core.report.Report`): The report with the
            sandbox instance.

    Returns:
        set[int]: The set of lines that have been executed in the student's
            code.
    """
    sandbox: Sandbox = report[TOOL_NAME]['sandbox']
    return sandbox.trace.lines - sandbox.trace.missing


def start_trace(report=MAIN_REPORT):
    """ Start tracing using the coverage module. """
    sandbox: Sandbox = report[TOOL_NAME]['sandbox']
    sandbox.tracer_style = 'coverage'


def stop_trace(report=MAIN_REPORT):
    """ Stop whatever tracing is going on. """
    sandbox: Sandbox = report[TOOL_NAME]['sandbox']
    sandbox.clear_tracer()


def check_coverage(report=MAIN_REPORT):
    """
    Checks that all the statements in the program have been executed.
    This function only works when a tracer_style has been set in the sandbox,
    or you are using an environment that automatically traces calls (e.g.,
    BlockPy).

    Args:
        report (Report): The Report to draw source code from; if not given,
            defaults to MAIN_REPORT.
    Returns:
        set[int]: If the source file was not parsed, None is returned.
            Otherwise, returnes the set of unexecuted lines.
        float: The ratio of unexected to total executable lines.
    """
    if not report[SOURCE_TOOL_NAME]['success']:
        return None, 0
    lines_executed = set(get_trace())
    # TODO: Why... does -1 get returned sometimes?
    if -1 in lines_executed:
        lines_executed.remove(-1)
    student_ast = report[SOURCE_TOOL_NAME]['ast']
    visitor = FindExecutableLines()
    visitor.visit(student_ast)
    lines_in_code = set(visitor.lines)
    unexecuted_lines = lines_in_code - lines_executed
    if lines_in_code:
        coverage_ratio = len(lines_executed)/len(lines_in_code)
    else:
        coverage_ratio = 0
    return unexecuted_lines, coverage_ratio


def clear_mocks(report=MAIN_REPORT):
    """ Reset mocked modules and functions to their defaults. """
    sandbox: Sandbox = report[TOOL_NAME]['sandbox']
    sandbox.clear_mocks()


def mock_function(function_name, new_version, report=MAIN_REPORT):
    """ Provide a custom version of the built-in function. """
    sandbox: Sandbox = report[TOOL_NAME]['sandbox']
    sandbox.mock_function(function_name, new_version)


def allow_function(function_name, report=MAIN_REPORT):
    """ Explicitly allow students to use the given function. """
    sandbox: Sandbox = report[TOOL_NAME]['sandbox']
    sandbox.allow_function(function_name)


def block_function(function_name, report=MAIN_REPORT):
    """ Cause an error if students call the given function. """
    sandbox: Sandbox = report[TOOL_NAME]['sandbox']
    sandbox.block_function(function_name)


def allow_module(module_name, report=MAIN_REPORT):
    """ Explicitly allow students to use the given module. """
    sandbox: Sandbox = report[TOOL_NAME]['sandbox']
    sandbox.allow_module(module_name)


def mock_module(module_name, new_version, friendly_name=None, report=MAIN_REPORT):
    """
    Provide an alternative version of the given module.

    Args:
        module_name (str): The importable name of the module.
        new_version (dict | :py:class:`pedal.sandbox.mocked.MockModule`): The
            new version of the module, either as a dictionary of fields/values
            or a fully created MockModule.
        friendly_name (str): The internal name to use to store the data for this
            module, accessible via Sandbox's `modules` field.
        report (:py:class:`pedal.core.report.Report`): The report with the
            sandbox instance.
    """
    sandbox: Sandbox = report[TOOL_NAME]['sandbox']
    sandbox.mock_module(module_name, new_version, friendly_name=friendly_name)


def block_module(module_name, report=MAIN_REPORT):
    """ Cause an error if students use the given module. """
    sandbox: Sandbox = report[TOOL_NAME]['sandbox']
    sandbox.block_module(module_name)


class CommandBlock:
    """
    Context Manager for creating instructor blocks of code that will
    be shown together to the student.

    TODO: What about named points where you can "rewind" the state to?
    """

    sandbox: Sandbox

    def __init__(self, report=MAIN_REPORT):
        self.sandbox = report[TOOL_NAME]['sandbox']

    def __enter__(self):
        self.sandbox.start_grouping_context()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.sandbox.stop_grouping_context()


# TODO: Deprecate these!
reset_output = clear_output



import ast

from pedal.types.definitions import (UnknownType, NumType, BoolType,
                                     TupleType, ListType, StrType,
                                     DictType, SetType, GeneratorType,
                                     DayType, TimeType, FunctionType, TYPE_STRINGS)


def merge_types(left, right):
    """

    Args:
        left:
        right:

    Returns:

    """
    # TODO: Check that lists/sets have the same subtypes
    if isinstance(left, (ListType, SetType, GeneratorType)):
        if left.empty:
            return right.subtype
        else:
            return left.subtype.clone()
    elif isinstance(left, TupleType):
        return left.subtypes + right.subtypes


def NumType_any(*x):
    """ Ignores all parameters to return a NumType """
    return NumType()


def StrType_any(*x):
    """ Ignores all parameters to return a StrType """
    return StrType()


def BoolType_any(*x):
    """ Ignores all parameters to return a BoolType """
    return BoolType()


def keep_left(left, right):
    """ Returns the left parameter """
    return left


def keep_right(left, right):
    """ Returns the right parameter """
    return right


# Maps the operations to their return types, based on the values.
# The *_any functions don't use the parameters, they return values unconditionally.
VALID_BINOP_TYPES = {
    ast.Add: {NumType: {NumType: NumType_any},
              StrType: {StrType: StrType_any},
              ListType: {ListType: merge_types},
              TupleType: {TupleType: merge_types}},
    ast.Sub: {NumType: {NumType: NumType_any},
              SetType: {SetType: merge_types}},
    ast.Div: {NumType: {NumType: NumType_any}},
    ast.FloorDiv: {NumType: {NumType: NumType_any}},
    ast.Mult: {NumType: {NumType: NumType_any,
                         StrType: StrType_any,
                         ListType: keep_right,
                         TupleType: keep_right},
               StrType: {NumType: StrType_any},
               ListType: {NumType: keep_left},
               TupleType: {NumType: keep_left}},
    ast.Pow: {NumType: {NumType: NumType_any}},
    # TODO: Should we allow old-fashioned string interpolation?
    # Currently, I vote no because it makes the code harder and is bad form.
    ast.Mod: {NumType: {NumType: NumType_any}},
    ast.LShift: {NumType: {NumType: NumType_any}},
    ast.RShift: {NumType: {NumType: NumType_any}},
    ast.BitOr: {NumType: {NumType: NumType_any},
                BoolType: {NumType: NumType_any,
                           BoolType: BoolType_any},
                SetType: {SetType: merge_types}},
    ast.BitXor: {NumType: {NumType: NumType_any},
                 BoolType: {NumType: NumType_any,
                            BoolType: BoolType_any},
                 SetType: {SetType: merge_types}},
    ast.BitAnd: {NumType: {NumType: NumType_any},
                 BoolType: {NumType: NumType_any,
                            BoolType: BoolType_any},
                 SetType: {SetType: merge_types}}
}

VALID_UNARYOP_TYPES = {
    ast.UAdd: {NumType: NumType},
    ast.USub: {NumType: NumType},
    ast.Invert: {NumType: NumType}
}


def are_types_equal(left, right, formal=False):
    """
    Determine if two types are equal.

    This could be more Polymorphic - move the code for each type into
    its respective class instead.

    Unknown types are not equal to anything, even other unknowns.

    Args:
        right:
        left:
        formal (bool): Whether the left argument is formal, indicating that it can accept
            type names.
    """
    if left is None or right is None:
        return False
    elif isinstance(left, UnknownType) or isinstance(right, UnknownType):
        return False
    elif not isinstance(left, type(right)):
        return False
    elif isinstance(left, (GeneratorType, ListType)):
        if left.empty or right.empty:
            return True
        else:
            return are_types_equal(left.subtype, right.subtype)
    elif isinstance(left, TupleType):
        if left.empty or right.empty:
            return True
        elif len(left.subtypes) != len(right.subtypes):
            return False
        else:
            for l, r in zip(left.subtypes, right.subtypes):
                if not are_types_equal(l, r):
                    return False
            return True
    elif isinstance(left, DictType):
        # print(left.empty, left.keys, left.literals, right)
        if not left.keys and not left.literals:
            return isinstance(right, DictType)
        # print("L", [literal.value for literal in left.literals], [v.singular_name
        #                                                          if not formal and not isinstance(v, FunctionType)
        #                                                          else TYPE_STRINGS[v.name]().singular_name
        #                                                          for v in left.values])
        # print("R", [literal.value for literal in right.literals], [v.singular_name for v in right.values])
        if left.empty or right.empty:
            return True
        elif left.literals is not None and right.literals is not None:
            if len(left.literals) != len(right.literals):
                return False
            else:
                for l, r in zip(left.literals, right.literals):
                    if not are_types_equal(l, r):
                        return False
                for l, r in zip(left.values, right.values):
                    if formal:
                        if isinstance(l, FunctionType) and l.name in TYPE_STRINGS:
                            l = TYPE_STRINGS[l.name]()
                        if isinstance(r, FunctionType) and r.name in TYPE_STRINGS:
                            r = TYPE_STRINGS[r.name]()
                    if not are_types_equal(l, r):
                        return False
                return True
        elif left.literals is not None or right.literals is not None:
            return False
        else:
            keys_equal = are_types_equal(left.keys, right.keys)
            values_equal = are_types_equal(left.values, right.values)
            return keys_equal and values_equal
    else:
        return True


ORDERABLE_TYPES = (NumType, BoolType, StrType, ListType, DayType, TimeType,
                   SetType, TupleType)
INDEXABLE_TYPES = (StrType, ListType, SetType, TupleType, DictType)



"""
Utilities for making type names more readable.
"""


def humanize_type(t: type) -> str:
    """
    Creates a human readable string name for a type.

    Args:
        t: Any type

    Returns:
        str: Human readable string
    """
    if hasattr(t, '__name__'):
        return t.__name__
    else:
        return str(t)


def humanize_types(types) -> str:
    """
    Convert a tuple of multiple types into a human readable representation.
    Also handles individual types by falling back to humanize_type.

    Args:
        types: A tuple of types, or an individual type.

    Returns:
        str: Human readable type name.
    """
    if isinstance(types, tuple):
        return ', '.join(humanize_type(t) for t in types)
    return humanize_type(types)



"""
Type definitions of builtin functions and modules.
"""

from pedal.types.definitions import (UnknownType, FunctionType,
                                     NumType, NoneType, BoolType,
                                     TupleType, ListType, StrType,
                                     FileType, DictType, ModuleType,
                                     SetType, DayType, TimeType,
                                     LiteralNum)

BUILTIN_MODULES = {
    'pprint': ModuleType('pprint',
                         fields={
                             'pprint': FunctionType(name='pprint', returns=NoneType())
                         }),
    'json': ModuleType('json',
                       fields={
                           'loads': FunctionType(name='loads', returns=UnknownType()),
                           'dumps': FunctionType(name='dumps', returns=StrType())
                       }),
    'random': ModuleType('random',
                         fields={
                             'randint': FunctionType(name='randint', returns=NumType())
                         }),
    'string': ModuleType('string',
                         fields={
                             'letters': StrType(empty=False),
                             'digits': StrType(empty=False),
                             'ascii_letters': StrType(empty=False),
                             'punctuation': StrType(empty=False),
                             'printable': StrType(empty=False),
                             'whitespace': StrType(empty=False),
                             'ascii_uppercase': StrType(empty=False),
                             'ascii_lowercase': StrType(empty=False),
                             'hexdigits': StrType(empty=False),
                             'octdigits': StrType(empty=False),
                         }),
    # TODO: Make this an extension in curriculum-ctvt
    'parking': ModuleType('parking',
                          fields={
                              'Time': FunctionType(name='Time', returns=TimeType()),
                              'now': FunctionType(name='now', returns=TimeType()),
                              'Day': FunctionType(name='Day', returns=DayType()),
                              'today': FunctionType(name='today', returns=DayType()),
                          }),
    'math': ModuleType('math',
                       fields={
                           'ceil': FunctionType(name='ceil', returns=NumType()),
                           'copysign': FunctionType(name='copysign', returns=NumType()),
                           'fabs': FunctionType(name='fabs', returns=NumType()),
                           'factorial': FunctionType(name='factorial', returns=NumType()),
                           'floor': FunctionType(name='floor', returns=NumType()),
                           'fmod': FunctionType(name='fmod', returns=NumType()),
                           'frexp': FunctionType(name='frexp', returns=NumType()),
                           'fsum': FunctionType(name='fsum', returns=NumType()),
                           'gcd': FunctionType(name='gcd', returns=NumType()),
                           'isclose': FunctionType(name='isclose', returns=BoolType()),
                           'isfinite': FunctionType(name='isfinite', returns=BoolType()),
                           'isinf': FunctionType(name='isinf', returns=BoolType()),
                           'isnan': FunctionType(name='isnan', returns=BoolType()),
                           'ldexp': FunctionType(name='ldexp', returns=NumType()),
                           'modf': FunctionType(name='modf', returns=NumType()),
                           'trunc': FunctionType(name='trunc', returns=NumType()),
                           'log': FunctionType(name='log', returns=NumType()),
                           'log1p': FunctionType(name='log1p', returns=NumType()),
                           'log2': FunctionType(name='log2', returns=NumType()),
                           'log10': FunctionType(name='log10', returns=NumType()),
                           'pow': FunctionType(name='pow', returns=NumType()),
                           'sqrt': FunctionType(name='sqrt', returns=NumType()),
                           'sin': FunctionType(name='sin', returns=NumType()),
                           'cos': FunctionType(name='cos', returns=NumType()),
                           'tan': FunctionType(name='tan', returns=NumType()),
                           'asin': FunctionType(name='asin', returns=NumType()),
                           'acos': FunctionType(name='acos', returns=NumType()),
                           'atan': FunctionType(name='atan', returns=NumType()),
                           'atan2': FunctionType(name='atan2', returns=NumType()),
                           'hypot': FunctionType(name='hypot', returns=NumType()),
                           'degrees': FunctionType(name='degrees', returns=NumType()),
                           'radians': FunctionType(name='radians', returns=NumType()),
                           'sinh': FunctionType(name='sinh', returns=NumType()),
                           'cosh': FunctionType(name='cosh', returns=NumType()),
                           'tanh': FunctionType(name='tanh', returns=NumType()),
                           'asinh': FunctionType(name='asinh', returns=NumType()),
                           'acosh': FunctionType(name='acosh', returns=NumType()),
                           'atanh': FunctionType(name='atanh', returns=NumType()),
                           'erf': FunctionType(name='erf', returns=NumType()),
                           'erfc': FunctionType(name='erfc', returns=NumType()),
                           'gamma': FunctionType(name='gamma', returns=NumType()),
                           'lgamma': FunctionType(name='lgamma', returns=NumType()),
                           'pi': NumType(),
                           'e': NumType(),
                           'tau': NumType(),
                           'inf': NumType(),
                           'nan': NumType(),
                       })
}


def get_builtin_module(name):
    """
    Given the name of the module, retrieve its TIFA representation.

    Args:
        name:

    Returns:

    """
    return BUILTIN_MODULES.get(name, UnknownType())


def _builtin_sequence_constructor(sequence_type):
    """
    Helper function for creating constructors for the Set and List types.
    These constructors use the subtype of the arguments.

    Args:
        sequence_type (Type): A function for creating new sequence types.
    """

    def sequence_call(tifa, function_type, callee, args, position):
        """

        Args:
            tifa:
            function_type:
            callee:
            args:
            position:

        Returns:

        """
        # TODO: Should inherit the emptiness too
        return_type = sequence_type(empty=True)
        if args:
            return_type.subtype = args[0].index(LiteralNum(0))
            return_type.empty = False
        return return_type

    return sequence_call


def _builtin_zip(tifa, function_type, callee, args, position):
    """
    Definition of the built-in zip function, which consumes a series of
    sequences and returns a list of tuples, with each tuple composed of the
    elements of the sequence paired (or rather, tupled) together.
    """
    if args:
        tupled_types = TupleType(subtypes=[])
        for arg in args:
            tupled_types.subtypes.append(arg.index(0))
        return ListType(tupled_types, empty=False)
    return ListType(empty=True)


# TODO: Exceptions

def get_builtin_function(name):
    """

    Args:
        name:

    Returns:

    """
    # Void Functions
    if name == "print":
        return FunctionType(name="print", returns=NoneType())
    # Math Functions
    elif name in ("int", "abs", "float", "len", "ord", "pow", "round", "sum"):
        return FunctionType(name=name, returns=NumType())
    # Boolean Functions
    elif name in ("bool", "all", "any", "isinstance"):
        return FunctionType(name=name, returns=BoolType())
    # String Functions
    elif name in ("str", 'chr', 'bin', 'repr', 'input'):
        return FunctionType(name=name, returns=StrType())
    # File Functions
    elif name == "open":
        return FunctionType(name="open", returns=FileType())
    # List Functions
    elif name == "map":
        return FunctionType(name="map", returns=ListType(empty=False))
    elif name == "list":
        return FunctionType(name="list",
                            definition=_builtin_sequence_constructor(ListType))
    # Set Functions
    elif name == "set":
        return FunctionType(name="set",
                            definition=_builtin_sequence_constructor(SetType))
    # Dict Functions
    elif name == "dict":
        return FunctionType(name="dict", returns=DictType())
    # Pass through
    elif name == "sorted":
        return FunctionType(name="sorted", returns='identity')
    elif name == "reversed":
        return FunctionType(name="reversed", returns='identity')
    elif name == "filter":
        return FunctionType(name="filter", returns='identity')
    # Special Functions
    elif name == "type":
        return FunctionType(name="type", returns=UnknownType())
    elif name == "range":
        return FunctionType(name="range", returns=ListType(NumType(), empty=False))
    elif name == "dir":
        return FunctionType(name="dir", returns=ListType(StrType(), empty=False))
    elif name == "max":
        return FunctionType(name="max", returns='element')
    elif name == "min":
        return FunctionType(name="min", returns='element')
    elif name == "zip":
        return FunctionType(name="zip", returns=_builtin_zip)
    elif name == "__import__":
        return FunctionType(name="__import__", returns=ModuleType())
    elif name == "globals":
        return FunctionType(name="globals",
                            returns=DictType(keys=StrType(),
                                             values=UnknownType(),
                                             empty=False))
    elif name in ("classmethod", "staticmethod"):
        return FunctionType(name=name, returns='identity')
    elif name in ("__name__",):
        return StrType()


# MatPlotLib Support
_PYPLOT_MODULE = ModuleType('pyplot', fields={
    'plot': FunctionType(name='plot', returns=NoneType()),
    'hist': FunctionType(name='hist', returns=NoneType()),
    'scatter': FunctionType(name='scatter', returns=NoneType()),
    'show': FunctionType(name='show', returns=NoneType()),
    'xlabel': FunctionType(name='xlabel', returns=NoneType()),
    'ylabel': FunctionType(name='ylabel', returns=NoneType()),
    'title': FunctionType(name='title', returns=NoneType()),
})

BUILTIN_MODULES['matplotlib'] = ModuleType('matplotlib',
                                           submodules={
                                               'pyplot': _PYPLOT_MODULE
                                           })


# TODO: Have to figure out how to setup this to be an extension.
#       Problem: It would be imported AFTER Tifa is run...
BUILTIN_MODULES['cisc108'] = ModuleType('cisc108', fields={
    'assert_equal': FunctionType(name='assert_equal', returns=NoneType()),
    'assert_type': FunctionType(name='assert_type', returns=NoneType()),
})






"""
Main TIFA visitor-based algorithm here.

TODO: JoinedStr
"""

import ast
# TODO: FileType, DayType, TimeType,
from pedal.core.commands import system_error
from pedal.tifa.tifa_core import TifaCore, TifaAnalysis
from pedal.types.definitions import (UnknownType,
                                     FunctionType, ClassType, InstanceType,
                                     NumType, NoneType, BoolType, TupleType,
                                     ListType, StrType, GeneratorType,
                                     DictType, ModuleType, SetType,
                                     LiteralNum, LiteralBool,
                                     LiteralNone, LiteralStr,
                                     LiteralTuple, Type)
from pedal.types.normalize import (get_pedal_type_from_json,
                                   get_pedal_literal_from_pedal_type,
                                   get_pedal_type_from_annotation,
                                   get_pedal_type_from_value)
from pedal.types.builtin import (get_builtin_module, get_builtin_function)
from pedal.types.operations import (are_types_equal,
                                    VALID_UNARYOP_TYPES, VALID_BINOP_TYPES,
                                    ORDERABLE_TYPES, INDEXABLE_TYPES)
from pedal.tifa.constants import TOOL_NAME
from pedal.tifa.contexts import NewPath, NewScope
from pedal.tifa.identifier import Identifier
from pedal.tifa.state import State
from pedal.tifa.feedbacks import (action_after_return, return_outside_function,
                                  write_out_of_scope, unconnected_blocks,
                                  iteration_problem, not_a_function,
                                  initialization_problem, unused_variable,
                                  overwritten_variable, iterating_over_non_list,
                                  iterating_over_empty_list, incompatible_types,
                                  parameter_type_mismatch, read_out_of_scope,
                                  type_changes, unnecessary_second_branch,
                                  recursive_call, multiple_return_types,
                                  possible_initialization_problem,
                                  incorrect_arity, else_on_loop_body,
                                  module_not_found, nested_function_definition,
                                  unused_returned_value, invalid_indexing)
from pedal.utilities.system import IS_PYTHON_39


class Tifa(TifaCore, ast.NodeVisitor):
    """
    TIFA subclass for traversing an AST and finding common issues.
    You can instantiate this class, manipulate settings, and then process
    some code or AST.
    """

    def process_code(self, code, filename=None, reset=True):
        """
        Processes the AST of the given source code to generate a report.

        Args:
            code (str): The Python source code
            filename (str): The filename of the source code (defaults to
                the submissions' main file).
            reset (bool): Whether or not to reset the results from the
                previous analysis before running this one.
        Returns:
            Report: The successful or successful report object
        """
        if reset or self.analysis is None:
            self.analysis = TifaAnalysis()
        filename = filename or self.report.submission.main_file
        self.line_offset = self.report.submission.line_offsets.get(filename, 0)

        # Attempt parsing - might fail!
        try:
            ast_tree = ast.parse(code, filename)
        except Exception as error:
            self.analysis.fail(error)
            system_error(TOOL_NAME, "Could not parse code: " + str(error),
                         report=self.report)
            return self.analysis
        # Attempt processing code - might fail!
        try:
            self.process_ast(ast_tree)
        except Exception as error:
            self.analysis.fail(error)
            system_error(TOOL_NAME, "Successfully parsed but could not "
                                    "process AST: " + str(error),
                         report=self.report)
        # Return whatever we got
        return self.analysis

    def process_ast(self, ast_tree):
        """
        Given an AST, actually performs the type and flow analyses to return a
        report.

        Args:
            ast_tree (AST): The AST object
        """
        self.reset()
        # Traverse every node
        self.visit(ast_tree)

        # Update analysis, finish out the current scope.
        self.analysis.variables = self.name_map
        self._finish_scope()

        # Collect top level variables
        self._collect_top_level_variables()

        return self.analysis

    def visit(self, node):
        """
        Process this node by calling its appropriate visit_*

        Args:
            node (AST): The node to visit
        Returns:
            Type: The type calculated during the visit.
        """
        # Start processing the node
        self.node_chain.append(node)
        self.ast_id += 1

        # Actions after return?
        if len(self.scope_chain) > 1:
            return_state = self.find_variable_scope("*return")
            if return_state.exists and return_state.in_scope:
                if return_state.state.set == "yes":
                    self._issue(action_after_return(self.locate(),
                                                    report=self.report))

        # No? All good, let's enter the node
        self.final_node = node
        result = ast.NodeVisitor.visit(self, node)
        # If a node failed to return something, return the UNKNOWN TYPE
        if result is None:
            result = UnknownType()
        self.analysis.node_types[node] = result

        # Pop the node out of the chain
        self.ast_id -= 1
        self.node_chain.pop()

        return result

    def _visit_nodes(self, nodes):
        """
        Visit all the nodes in the given list.

        Args:
            nodes (list): A list of values, of which any AST nodes will be
                          visited.
        """
        for node in nodes:
            if isinstance(node, ast.AST):
                self.visit(node)

    @staticmethod
    def walk_targets(targets, target_type, walker):
        """
        Iterate through the targets and call the given function on each one.

        Args:
            targets (list of Ast nodes): A list of potential targets to be
                                         traversed.
            target_type (Type): The given type to be unraveled and applied to the
                         targets.
            walker (Ast Node, Type -> None): A function that will process
                                             each target and unravel the type.
        """
        for target in targets:
            walker(target, target_type)

    def _walk_target(self, target, target_type):
        """
        Recursively apply the type to the target

        Args:
            target (Ast): The current AST node to process
            target_type (Type): The type to apply to this node
        """
        if isinstance(target, ast.Name):
            self.store_iter_variable(target.id, target_type, self.locate(target))
            return target.id
        elif isinstance(target, (ast.Tuple, ast.List)):
            result = None
            for i, elt in enumerate(target.elts):
                elt_type = target_type.iterate(LiteralNum(i))
                potential_name = self._walk_target(elt, elt_type)
                if potential_name is not None and result is None:
                    result = potential_name
            return result

    # TODO: Properly handle assignments with subscripts
    def assign_target(self, target, target_type):
        """
        Assign the type to the target, handling all kinds of assignment
        statements, including Names, Tuples/Lists, Subscripts, and
        Attributes.

        Args:
            target (ast.AST): The target AST Node.
            target_type (Type): The TIFA type.

        Returns:

        """
        if isinstance(target, ast.Name):
            self.store_variable(target.id, target_type)
        elif isinstance(target, (ast.Tuple, ast.List)):
            for i, elt in enumerate(target.elts):
                elt_type = target_type.iterate(LiteralNum(i))
                self.assign_target(elt, elt_type)
        elif isinstance(target, ast.Subscript):
            left_hand_type = self.visit(target.value)
            if isinstance(left_hand_type, ListType):
                # TODO: Handle updating value in list
                pass
            elif isinstance(left_hand_type, DictType):
                # TODO: Update this for Python 3.9, now that Slice notation has changed
                if not isinstance(target.slice, ast.Index):
                    # TODO: Can't subscript a dictionary assignment
                    return None
                literal = self.get_literal(target.slice.value)
                if not literal:
                    key_type = self.visit(target.slice.value)
                    left_hand_type.empty = False
                    left_hand_type.keys = [key_type.clone()]
                    left_hand_type.values = [target_type.clone()]
                elif left_hand_type.literals:
                    original_type = left_hand_type.has_literal(literal)
                    if not original_type:
                        left_hand_type.update_key(literal, target_type.clone())
                    elif not are_types_equal(original_type, target_type):
                        # TODO: Fix "Dictionary" to be the name of the variable
                        self._issue(type_changes(self.locate(), 'Dictionary', original_type, target_type))
        elif isinstance(target, ast.Attribute):
            left_hand_type = self.visit(target.value)
            if isinstance(left_hand_type, InstanceType):
                left_hand_type.add_attr(target.attr, target_type)
            # TODO: Otherwise we attempted to assign to a non-instance
            # TODO: Handle minor type changes (e.g., appending to an inner list)

    def _visit_collection_loop(self, node):
        was_empty = False
        # Handle the iteration list
        iter_list = node.iter
        iter_list_name = None
        if isinstance(iter_list, ast.Name):
            iter_list_name = iter_list.id
            if iter_list_name == "___":
                self._issue(unconnected_blocks(self.locate(iter_list), report=self.report))
            state = self.iterate_variable(iter_list_name, self.locate(iter_list))
            iter_type = state.type
        else:
            iter_type = self.visit(iter_list)

        if iter_type.is_empty():
            # TODO: It should check if its ONLY ever iterating over an empty list.
            # For now, only reports if we are NOT in a function
            was_empty = True
            if len(self.scope_chain) == 1:
                self._issue(iterating_over_empty_list(self.locate(iter_list), iter_list_name))

        if not isinstance(iter_type, INDEXABLE_TYPES):
            self._issue(iterating_over_non_list(self.locate(iter_list), iter_list_name, report=self.report))

        iter_subtype = iter_type.iterate(LiteralNum(0))

        # Handle the iteration variable
        iter_variable_name = self._walk_target(node.target, iter_subtype)

        # Check that the iteration list and variable are distinct
        if iter_variable_name and iter_list_name:
            if iter_variable_name == iter_list_name:
                self._issue(iteration_problem(self.locate(node.target), iter_variable_name))

        return was_empty

    def visit_AnnAssign(self, node):
        """
        TODO: Implement!
        """
        # Name, Attribute, or SubScript
        target = node.target
        # Type
        annotation = node.annotation
        # Optional assigned value
        value = node.value
        # 0 or 1, with 1 indicating pure names (not expressions)
        simple = node.simple
        # If it's a class attribute, then build up the type!
        if simple:
            self.visit(annotation)
            annotation = get_pedal_type_from_annotation(annotation, self)
            current_scope = self.scope_chain[0]
            if current_scope in self.class_scopes:
                # TODO: Treat it as a different kind of ClassType? TypedDict?
                self.class_scopes[current_scope].add_attr(target.id, annotation)

    def visit_Expr(self, node):
        """
        Any expression being used as a statement.

        Args:
            node (AST): An Expr node

        Returns:

        """
        value = self.visit(node.value)
        if isinstance(node.value, ast.Call) and not isinstance(value, NoneType):
            # TODO: Helper function to get name with title ("append method")
            if isinstance(node.value.func, ast.Name):
                call_type = 'function'
            else:
                call_type = 'method'
            name = self.identify_caller(node.value)
            self._issue(unused_returned_value(self.locate(), name,
                                              call_type, value))

    def visit_Assign(self, node):
        """
        Simple assignment statement:
        __targets__ = __value__

        Args:
            node (AST): An Assign node
        Returns:
            None
        """
        # Handle value
        value_type = self.visit(node.value)
        # Handle targets
        self._visit_nodes(node.targets)

        self.walk_targets(node.targets, value_type, self.assign_target)

    def visit_AugAssign(self, node):
        """

        Args:
            node:

        Returns:

        """
        # Handle value
        right = self.visit(node.value)
        # Handle target
        left = self.visit(node.target)
        # Target is always a Name, Subscript, or Attribute
        name = self.identify_caller(node.target)

        # Handle operation
        self.load_variable(name)
        if isinstance(left, UnknownType) or isinstance(right, UnknownType):
            return UnknownType()
        elif type(node.op) in VALID_BINOP_TYPES:
            op_lookup = VALID_BINOP_TYPES[type(node.op)]
            if type(left) in op_lookup:
                op_lookup = op_lookup[type(left)]
                if type(right) in op_lookup:
                    op_lookup = op_lookup[type(right)]
                    result_type = op_lookup(left, right)
                    self.assign_target(node.target, result_type)
                    return result_type
        self._issue(incompatible_types(self.locate(), node.op, left, right, report=self.report))

    def visit_Attribute(self, node):
        """

        Args:
            node:

        Returns:

        """
        # Handle value
        value_type = self.visit(node.value)
        self.check_common_bad_lookups(value_type, node.attr, node.value)
        # Handle ctx
        # TODO: Handling contexts
        # Handle attr
        result = value_type.load_attr(node.attr)
        return result

    def visit_BinOp(self, node):
        """

        Args:
            node:

        Returns:

        """
        # Handle left and right
        left = self.visit(node.left)
        right = self.visit(node.right)

        # Handle operation
        if isinstance(left, UnknownType) or isinstance(right, UnknownType):
            return UnknownType()
        elif type(node.op) in VALID_BINOP_TYPES:
            op_lookup = VALID_BINOP_TYPES[type(node.op)]
            if type(left) in op_lookup:
                op_lookup = op_lookup[type(left)]
                if type(right) in op_lookup:
                    op_lookup = op_lookup[type(right)]
                    return op_lookup(left, right)
        self._issue(incompatible_types(self.locate(), node.op, left, right, report=self.report))
        return UnknownType()

    def visit_Bool(self, node):
        """
        Visit a constant boolean value.

        Args:
            node (ast.AST): The boolean value Node.

        Returns:
            Type: A Bool type.
        """
        return BoolType()

    def visit_BoolOp(self, node):
        """

        Args:
            node:

        Returns:

        """
        # Handle left and right
        values = []
        for value in node.values:
            values.append(self.visit(value))

        # TODO: Truthiness is not supported! Probably need a Union type
        # TODO: Literals used as truthy value

        # Handle operation
        return BoolType()

    def visit_Call(self, node):
        """

        Args:
            node:

        Returns:

        """
        # Handle func part (Name or Attribute)
        function_type = self.visit(node.func)
        # TODO: Need to grab the actual type in some situations
        callee = self.identify_caller(node)

        # Handle args
        arguments = [self.visit(arg) for arg in node.args] if node.args else []

        # Check special common mistakes

        # TODO: Handle keywords
        # TODO: Handle star args
        # TODO: Handle kwargs
        if isinstance(function_type, FunctionType):
            # Test if we have called this definition before
            if function_type.definition not in self.definition_chain:
                self.definition_chain.append(function_type.definition)
                # Function invocation
                result = function_type.definition(self, function_type, callee,
                                                  arguments, self.locate())
                self.definition_chain.pop()
                return result
            else:
                self._issue(recursive_call(self.locate(), callee, report=self.report))
        elif isinstance(function_type, ClassType):
            constructor = function_type.get_constructor().definition
            self.definition_chain.append(constructor)
            result = constructor(self, constructor, callee, arguments, self.locate())
            self.definition_chain.pop()
            if '__init__' in function_type.fields:
                initializer = function_type.fields['__init__']
                if isinstance(initializer, FunctionType):
                    self.definition_chain.append(initializer)
                    initializer.definition(self, initializer, result, [result] + arguments, self.locate())
                    self.definition_chain.pop()
            return result
        elif isinstance(function_type, (NumType, StrType, BoolType, NoneType)):
            self._issue(not_a_function(self.locate(), callee, function_type, report=self.report))
        return UnknownType()

    def visit_ClassDef(self, node):
        """

        Args:
            node:
        """
        class_name = node.name
        new_class_type = ClassType(class_name)
        self.store_variable(class_name, new_class_type)
        # TODO: Define a new scope definition that executes the body
        # TODO: find __init__, execute that
        # TODO: handle Record subclasses
        definitions_scope = self.scope_chain[:]
        class_scope = NewScope(self, definitions_scope, class_type=new_class_type)
        with class_scope:
            self.generic_visit(node)

    def visit_Compare(self, node):
        """

        Args:
            node:

        Returns:

        """
        # Handle left and right
        left = self.visit(node.left)
        comparators = [self.visit(compare) for compare in node.comparators]

        # Handle ops
        for op, right in zip(node.ops, comparators):
            if isinstance(op, (ast.Eq, ast.NotEq, ast.Is, ast.IsNot)):
                continue
            elif isinstance(op, (ast.Lt, ast.LtE, ast.GtE, ast.Gt)):
                if are_types_equal(left, right):
                    if isinstance(left, ORDERABLE_TYPES):
                        continue
            elif isinstance(op, (ast.In, ast.NotIn)):
                if isinstance(right, INDEXABLE_TYPES):
                    continue
            self._issue(incompatible_types(self.locate(), op, left, right, report=self.report))
        return BoolType()

    def visit_comprehension(self, node):
        """

        Args:
            node:
        """
        self._visit_collection_loop(node)
        # Handle ifs, unless they're blank (None in Skulpt :)
        if node.ifs:
            self.visit_statements(node.ifs)

    def visit_Dict(self, node):
        """
        Three types of dictionaries
        - empty
        - uniform type
        - record
        TODO: Handle records appropriately
        """
        result_type = DictType()
        if not node.keys:
            result_type.empty = True
        else:
            result_type.empty = False
            all_literals = True
            keys, values, literals = [], [], []
            for key, value in zip(node.keys, node.values):
                literal = self.get_literal(key)
                key, value = self.visit(key), self.visit(value)
                values.append(value)
                keys.append(key)
                if literal is not None:
                    literals.append(literal)
                else:
                    all_literals = False

            if all_literals:
                result_type.literals = literals
                result_type.values = values
            else:
                result_type.keys = node.keys[0]
                result_type.values = node.values[0]
        return result_type

    def visit_DictComp(self, node):
        """

        Args:
            node:

        Returns:

        """
        # TODO: Handle comprehension scope
        for generator in node.generators:
            self.visit(generator)
        keys = self.visit(node.key)
        values = self.visit(node.value)
        return DictType(keys=keys, values=values)

    def visit_For(self, node):
        """

        Args:
            node:
        """
        was_empty = self._visit_collection_loop(node)
        # Handle the bodies
        #if not was_empty:
        #this_path_id = self.path_chain[0]
        #non_empty_path = NewPath(self, this_path_id, "f")
        #with non_empty_path:
        self.visit_statements(node.body)
        self.visit_statements(node.orelse)

    def visit_FunctionDef(self, node):
        """

        Args:
            node:

        Returns:

        """
        # Name
        function_name = node.name
        position = self.locate()
        definitions_scope = self.scope_chain[:]

        def definition(tifa, call_type, call_name, parameters, call_position):
            """

            Args:
                tifa:
                call_type:
                call_name:
                parameters:
                call_position:

            Returns:

            """
            function_scope = NewScope(self, definitions_scope)
            with function_scope:
                # Process arguments
                args = node.args.args
                if len(args) != len(parameters):
                    self._issue(incorrect_arity(self.locate(), function_name, report=self.report))
                # TODO: Handle special types of parameters
                for arg, parameter in zip(args, parameters):
                    name = arg.arg
                    if arg.annotation:
                        self.visit(arg.annotation)
                        annotation = get_pedal_type_from_annotation(arg.annotation, self)
                        # TODO: Use parameter information to "fill in" empty lists
                        if isinstance(parameter, ListType) and isinstance(annotation, ListType):
                            if isinstance(parameter.subtype, UnknownType):
                                parameter.subtype = annotation.subtype
                        # TODO: Check that arg.type and parameter type match!
                        if not are_types_equal(annotation, parameter, True):
                            self._issue(parameter_type_mismatch(self.locate(), name, annotation, parameter))
                    if parameter is not None:
                        parameter = parameter.clone_mutably()
                        self.create_variable(name, parameter, position)
                # Too many arguments
                if len(args) < len(parameters):
                    for undefined_parameter in parameters[len(args):]:
                        self.create_variable(name, UnknownType(), position)
                # Not enough arguments
                if len(args) > len(parameters):
                    for arg in args[len(parameters):]:
                        if arg.annotation:
                            self.visit(arg.annotation)
                            annotation = get_pedal_type_from_annotation(arg.annotation, self)
                        else:
                            annotation = UnknownType()
                        self.create_variable(arg.arg, annotation, position)
                self.visit_statements(node.body)
                return_state = self.find_variable_scope("*return")
                return_value = NoneType()
                # TODO: Figure out if we are not returning something when we should
                # If the pseudo variable exists, we load it and get its type
                if return_state.exists and return_state.in_scope:
                    return_state = self.load_variable("*return", call_position)
                    return_value = return_state.type
                    if node.returns:
                        # self.visit(node.returns)
                        returns = get_pedal_type_from_annotation(node.returns, self)
                        if not are_types_equal(return_value, returns, True):
                            self._issue(multiple_return_types(return_state.position,
                                                              returns.precise_description(),
                                                              return_value.precise_description(),
                                                              report=self.report))
            return return_value

        function = FunctionType(definition=definition, name=function_name)
        self.store_variable(function_name, function)

        if len(self.node_chain) > 2:
            self._issue(nested_function_definition(self.locate(), function_name))

        return function

    def visit_GeneratorExp(self, node):
        """

        Args:
            node:

        Returns:

        """
        # TODO: Handle comprehension scope
        for generator in node.generators:
            self.visit(generator)
        return GeneratorType(self.visit(node.elt))

    def visit_If(self, node):
        """

        Args:
            node:
        """
        # Visit the conditional
        self.visit(node.test)

        if len(node.orelse) == 1 and isinstance(node.orelse[0], ast.Pass):
            self._issue(unnecessary_second_branch(self.locate()))
        elif len(node.body) == 1 and isinstance(node.body[0], ast.Pass):
            if node.orelse:
                self._issue(unnecessary_second_branch(self.locate()))

        # Visit the bodies
        this_path_id = self.path_chain[0]
        if_path = NewPath(self, this_path_id, "i")
        with if_path:
            for statement in node.body:
                self.visit(statement)
        else_path = NewPath(self, this_path_id, "e")
        with else_path:
            for statement in node.orelse:
                self.visit(statement)

        # TODO: Unconditional return

        # Combine two paths into one
        # Check for any names that are on the IF path
        self.merge_paths(this_path_id, if_path.id, else_path.id)

    def visit_IfExp(self, node):
        """

        Args:
            node:

        Returns:

        """
        # Visit the conditional
        self.visit(node.test)

        # Visit the body
        body = self.visit(node.body)

        # Visit the orelse
        orelse = self.visit(node.orelse)

        if are_types_equal(body, orelse):
            return body

        # TODO: Union type?
        return UnknownType()

    def visit_Import(self, node):
        """

        Args:
            node:
        """
        # Handle names
        for alias in node.names:
            asname = alias.asname or alias.name
            module_type = self.load_module(alias.name)
            self.store_variable(asname, module_type)

    def visit_ImportFrom(self, node):
        """

        Args:
            node:
        """
        # Handle names
        for alias in node.names:
            if node.module is None:
                asname = alias.asname or alias.name
                module_type = self.load_module(alias.name)
            else:
                module_name = node.module
                asname = alias.asname or alias.name
                module_type = self.load_module(module_name)
            name_type = module_type.load_attr(alias.name)
            self.store_variable(asname, name_type)

    def visit_Lambda(self, node):
        """

        Args:
            node:

        Returns:

        """
        # Name
        position = self.locate()
        definitions_scope = self.scope_chain[:]

        def definition(tifa, call_type, call_name, parameters, call_position):
            """

            Args:
                tifa:
                call_type:
                call_name:
                parameters:
                call_position:

            Returns:

            """
            function_scope = NewScope(self, definitions_scope)
            with function_scope:
                # Process arguments
                args = node.args.args
                if len(args) != len(parameters):
                    self._issue(incorrect_arity(position, "lambda", report=self.report))
                # TODO: Handle special types of parameters
                for arg, parameter in zip(args, parameters):
                    name = arg.arg
                    if parameter is not None:
                        parameter = parameter.clone_mutably()
                        self.store_variable(name, parameter, position)
                if len(args) < len(parameters):
                    for undefined_parameter in parameters[len(args):]:
                        self.store_variable(name, UnknownType(), position)
                return_value = self.visit(node.body)
            return return_value

        return FunctionType(definition=definition)

    def visit_List(self, node):
        """

        Args:
            node:

        Returns:

        """
        result_type = ListType()
        if node.elts:
            result_type.empty = False
            # TODO: confirm homogenous subtype
            for elt in node.elts:
                result_type.subtype = self.visit(elt)
        else:
            result_type.empty = True
        return result_type

    def visit_ListComp(self, node):
        """

        Args:
            node:

        Returns:

        """
        # TODO: Handle comprehension scope
        for generator in node.generators:
            self.visit(generator)
        return ListType(self.visit(node.elt))

    def visit_NameConstant(self, node):
        """

        Args:
            node:

        Returns:

        """
        value = node.value
        if isinstance(value, bool):
            return BoolType()
        else:
            return NoneType()

    def visit_Name(self, node):
        """

        Args:
            node:

        Returns:

        """
        name = node.id
        if name == "___":
            self._issue(unconnected_blocks(self.locate()))
        if isinstance(node.ctx, ast.Load):
            if name == "True" or name == "False":
                return BoolType()
            elif name == "None":
                return NoneType()
            else:
                variable = self.find_variable_scope(name)
                builtin = get_builtin_function(name)
                if not variable.exists and builtin:
                    return builtin
                else:
                    state = self.load_variable(name)
                    return state.type
        else:
            variable = self.find_variable_scope(name)
            if variable.exists:
                return variable.state.type
            else:
                return UnknownType()

    def visit_Num(self, node):
        """

        Args:
            node:

        Returns:

        """
        return NumType()

    def visit_Constant(self, node) -> Type:
        """ Handle new 3.8's Constant node """
        return get_pedal_type_from_value(node.value)

    def visit_Return(self, node):
        """

        Args:
            node:
        """
        if len(self.scope_chain) == 1:
            self._issue(return_outside_function(self.locate(), report=self.report))
        # TODO: Unconditional return inside loop
        if node.value is not None:
            self.return_variable(self.visit(node.value))
        else:
            self.return_variable(NoneType())

    def visit_SetComp(self, node):
        """

        Args:
            node:

        Returns:

        """
        # TODO: Handle comprehension scope
        for generator in node.generators:
            self.visit(generator)
        return SetType(self.visit(node.elt))

    def visit_statements(self, nodes):
        """

        Args:
            nodes:

        Returns:

        """
        # TODO: Check for pass in the middle of a series of statement
        if any(isinstance(node, ast.Pass) for node in nodes):
            pass
        return [self.visit(statement) for statement in nodes]

    def visit_Str(self, node):
        """

        Args:
            node:

        Returns:

        """
        if node.s == "":
            return StrType(True)
        else:
            return StrType(False)

    def visit_JoinedStr(self, node):
        values = [self.visit(expr) for expr in node.values]
        # The result will be all StrType
        return StrType(empty=all(n.empty for n in values))

    def visit_FormattedValue(self, node):
        value = self.visit(node.value)
        if isinstance(value, StrType):
            return value
        else:
            return StrType(empty=False)

    def visit_Subscript(self, node):
        """

        Args:
            node:

        Returns:

        """
        # Handle value
        value_type = self.visit(node.value)
        # Handle slice
        if IS_PYTHON_39 and isinstance(node.slice, ast.Tuple):
            # TODO: Do something about extslices (especially since students stumble into this accidentally)
            pass
        elif not IS_PYTHON_39 and isinstance(node.slice, ast.ExtSlice):
            # TODO: Do something about extslices (especially since students stumble into this accidentally)
            pass
        elif isinstance(node.slice, ast.Slice):
            self.visit_Slice(node.slice)
            return value_type
        else:
            if IS_PYTHON_39:
                slice = node.slice
            else:
                slice = node.slice.value
            literal = self.get_literal(slice)
            if literal is None:
                literal = get_pedal_literal_from_pedal_type(self.visit(slice))
            result = value_type.index(literal)
            # TODO: Is this sufficient? Maybe we should be throwing?
            if isinstance(result, UnknownType):
                self._issue(invalid_indexing(self.locate(), value_type,
                                             literal.type()))
            else:
                return result

    def visit_Slice(self, node):
        """ Handles a slice by visiting its components; cannot return a value
        because the slice is always the same type as its value, which is
        not available on the Slice node itself. """
        if node.lower is not None:
            self.visit(node.lower)
        if node.upper is not None:
            self.visit(node.upper)
        if node.step is not None:
            self.visit(node.step)

    def visit_Tuple(self, node) -> TupleType:
        """ Handle Tuple literal """
        result_type = TupleType()
        if not node.elts:
            result_type.empty = True
            result_type.subtypes = []
        else:
            result_type.empty = False
            # TODO: confirm homogenous subtype
            result_type.subtypes = [self.visit(elt) for elt in node.elts]
        return result_type

    def visit_UnaryOp(self, node):
        """

        Args:
            node:

        Returns:

        """
        # Handle operand
        operand = self.visit(node.operand)

        if isinstance(node.op, ast.Not):
            return BoolType()
        elif isinstance(operand, UnknownType):
            return UnknownType()
        elif type(node.op) in VALID_UNARYOP_TYPES:
            op_lookup = VALID_UNARYOP_TYPES[type(node.op)]
            if type(operand) in op_lookup:
                return op_lookup[type(operand)]()
        return UnknownType()

    def visit_While(self, node):
        """

        Args:
            node:
        """
        # Visit conditional
        self.visit(node.test)

        # Visit the bodies
        this_path_id = self.path_id
        # One path is that we never enter the body
        empty_path = NewPath(self, this_path_id, "e")
        with empty_path:
            pass
        # Another path is that we loop through the body and check the test again
        body_path = NewPath(self, this_path_id, "w")
        with body_path:
            for statement in node.body:
                self.visit(statement)
            # Revisit conditional
            self.visit(node.test)
        # If there's else bodies (WEIRD) then we should check them afterwards
        if node.orelse:
            self._issue(else_on_loop_body(self.locate()))
            for statement in node.orelse:
                self.visit(statement)

        # Combine two paths into one
        # Check for any names that are on the IF path
        self.merge_paths(this_path_id, body_path.id, empty_path.id)

    def visit_With(self, node):
        """

        Args:
            node:
        """
        for item in node.items:
            type_value = self.visit(item.context_expr)
            self.visit(item.optional_vars)
            self._walk_target(item.optional_vars, type_value)
        # Handle the bodies
        self.visit_statements(node.body)



"""
Human-readable names for operators.
"""
import ast

OPERATION_DESCRIPTION = {
    ast.Pow: "an exponent",
    ast.Add: "an addition",
    ast.Mult: "a multiplication",
    ast.Sub: "a subtraction",
    ast.Div: "a division",
    ast.FloorDiv: "a division",
    ast.Mod: "a modulo",
    ast.LShift: "a left shift",
    ast.RShift: "a right shift",
    ast.BitOr: "a bit or",
    ast.BitAnd: "a bit and",
    ast.BitXor: "a bit xor",
    ast.And: "an and",
    ast.Or: "an or",
    ast.Eq: "an ==",
    ast.NotEq: "a !=",
    ast.Lt: "a <",
    ast.LtE: "a <=",
    ast.Gt: "a >",
    ast.GtE: "a >=",
    ast.Is: "an is",
    ast.IsNot: "an is not",
    ast.In: "an in",
    ast.NotIn: "an not in",
}

COMPARE_OP_NAMES = {
    "==": "Eq",
    "<": "Lt",
    "<=": "Lte",
    ">=": "Gte",
    ">": "Gt",
    "!=": "NotEq",
    "is": "Is",
    "is not": "IsNot",
    "in": "In",
    "not in": "NotIn",
}

BOOL_OP_NAMES = {"and": "And", "or": "Or"}

BIN_OP_NAMES = {
    "+": "Add",
    "-": "Sub",
    "*": "Mult",
    "/": "Div",
    "//": "FloorDiv",
    "%": "Mod",
    "**": "Pow",
    ">>": "LShift",
    "<<": "RShift",
    "|": "BitOr",
    "^": "BitXor",
    "&": "BitAnd",
    "@": "MatMult",
}

UNARY_OP_NAMES = {
    # "+": "UAdd",
    # "-": "USub",
    "not": "Not",
    "~": "Invert",
}



""" Utilities related to file handling """
import os


def normalize_path(first: str, second: str):
    """
    If the `first` path is relative, creates an absolute path relative to the `second` folder.

    Args:
        first: The path to normalize
        second: The path to normalize in regards to.

    Returns:
        str: A new path.
    """
    if os.path.isabs(first):
        return first
    else:
        return os.path.join(os.path.dirname(second), first)



"""
Environments are a collection of defaults, setups, and overrides that make
Pedal adapt better to a given autograding platform (e.g., BlockPy, WebCAT,
GradeScope). They are meant to streamline common configuration.
"""

__all__ = ['Environment']

from pedal.core.report import MAIN_REPORT
from pedal.core.submission import Submission


class Environment:
    """
    Abstract Environment class, meant to be subclassed by the environment to
    help simplify configuration. Technically doesn't need to do anything.
    Creating an instance of an environment will automatically clear out the
    existing contents of the report.

    Args:
        main_file (str): The filename of the main file.
        main_code (str): The actual code of the main file.
        files (dict[str, str]): A list of filenames mapped to their contents.
    """
    def __init__(self, files=None, main_file='answer.py', main_code=None,
                 user=None, assignment=None, course=None, execution=None,
                 instructor_file='instructor.py', report=MAIN_REPORT):
        self.report = report
        report.clear()
        # Setup any code given as the submission.
        if isinstance(files, Submission):
            self.submission = files
        else:
            load_error = None
            if files is None:
                if main_code is None:
                    if main_file is None:
                        raise ValueError("files, main_code, and main_file cannot all be None.")
                    else:
                        main_code = self.load_main(main_file)
                        if isinstance(main_code, Exception):
                            load_error = main_code
                            main_code = ""
                files = {main_file: main_code}
            else:
                if main_file is not None:
                    if main_code is None:
                        main_code = files[main_file]
                    if main_file not in files:
                        files[main_file] = main_code
            self.submission = Submission(files, main_file, main_code,
                                         user, assignment, course, execution,
                                         instructor_file, load_error=load_error)
        # Contextualize report
        self.report.contextualize(self.submission)
        self.fields = {}

    def load_main(self, main_file):
        with open(main_file) as opened:
            return opened.read()

    def __iter__(self):
        return self.fields.items()

    def __getattr__(self, key):
        try:
            return self.fields[key]
        except KeyError:
            raise AttributeError(key)



"""
Utilities for handling English Language rules.

"""


def add_indefinite_article(phrase):
    """
    Given a word, choose the appropriate indefinite article (e.g., "a" or "an")
    and add it the front of the word. Matches the original word's
    capitalization.

    Args:
        phrase (str): The phrase to get the appropriate definite article
            for.

    Returns:
        str: Either "an" or "a".
    """
    if phrase[0] in "aeiou":
        return "an "+phrase
    elif phrase[0] in "AEIOU":
        return "An "+phrase
    elif phrase[0].lower() == phrase[0]:
        return "a "+phrase
    else:
        return "A "+phrase


def escape_curly_braces(result):
    """ Replaces all occurrences of curly braces with double curly braces. """
    return result.replace("{", "{{").replace("}", "}}")


def safe_repr(obj, short=False, max_length=80) -> str:
    """
    Create a string representation of this object using :py:func:`repr`. If the
    object doesn't have a repr defined, then falls back onto the default
    :py:func:`object.__repr__`. Finally, if ``short`` is true, then the string
    will be truncated to the max_length value.

    Args:
        obj (Any): Any object to repr.
        short (bool): Whether or not to truncate to the max_length.
        max_length (int): How many maximum characters to use, if short is True.
    """
    try:
        result = repr(obj)
    except Exception:
        result = object.__repr__(obj)
    if short and len(result) >= max_length:
        result = result[:max_length] + ' [truncated]...'
    result = result
    return result


def chomp(text):
    """ Removes the trailing newline character, if there is one. """
    if not text:
        return text
    if text[-2:] == '\n\r':
        return text[:-2]
    if text[-1] == '\n':
        return text[:-1]
    return text



from pedal.core.commands import compliment, explain, gently
from pedal.core.report import MAIN_REPORT
from pedal.sandbox.commands import run
from pedal.assertions.runtime import *
from pedal.assertions.static import ensure_function


class QuestionGrader:
    def _get_functions_with_filter(self, filter='grade_'):
        return [getattr(self, method_name) for method_name in dir(self)
                if method_name.startswith(filter) and
                callable(getattr(self, method_name))]

    def _test(self, question):
        methods = self._get_functions_with_filter()
        for method in methods:
            method(question)


class FunctionGrader(QuestionGrader):
    """

    """
    MAX_POINTS = 10
    DEFINITION_POINTS = 3
    COMPONENTS_POINTS = 1
    MAX_COMPONENTS_POINTS = 2
    UNIT_TEST_TYPE_POINTS = None
    UNIT_TEST_VALUE_POINTS = None
    UNIT_TEST_TOTAL_POINTS = 5
    UNIT_TEST_TYPE_RATIO = .5
    UNIT_TEST_COMPLETION_POINTS = 2

    def __init__(self, function_name, signature, tests):
        super().__init__()
        self.function_name = function_name
        self.signature = signature
        self.tests = tests
        self.points = 0

    def _test(self, question):
        defined = self.grade_definition(question)

        if not defined:
            return self.report_status(question)

        self.grade_components(question)

        passed_tests = self.grade_unit_tests(question)
        if not passed_tests:
            return self.report_status(question)

        self.report_success(question)

    def report_status(self, question):
        """

        Args:
            question:
        """
        pass

    def report_success(self, question):
        """

        Args:
            question:
        """
        question.answer()

    def grade_definition(self, question):
        """

        Args:
            question:

        Returns:

        """
        self.student = run()

        if not ensure_function(self.function_name, *self.signature):
            gently("Function not defined")
            return False

        if self.student.exception:
            return False
        if not assertHasFunction(self.student, self.function_name):
            gently("Function defined incorrectly")
            return False

        self.points += self.DEFINITION_POINTS
        return True

    def grade_components(self, question):
        """

        Args:
            question:
        """
        self.component_points = 0
        components = self._get_functions_with_filter('grade_component_')
        for component in components:
            component(question)
        self.component_points = min(self.component_points, self.MAX_COMPONENTS_POINTS)
        self.points += self.component_points

    def assertEqual(self, *parameters):
        """

        Args:
            *parameters:

        Returns:

        """
        return assertEqual(*parameters)

    def grade_unit_tests(self, question):
        """

        Args:
            question:

        Returns:

        """
        all_good = True
        if self.UNIT_TEST_TOTAL_POINTS is None:
            TYPE_POINT_ADD = self.UNIT_TEST_TYPE_POINTS
            VALUE_POINT_ADD = self.UNIT_TEST_VALUE_POINTS
        else:
            ratio = self.UNIT_TEST_TYPE_RATIO
            TYPE_POINT_ADD = (self.UNIT_TEST_TOTAL_POINTS / len(self.tests) * (ratio))
            VALUE_POINT_ADD = (self.UNIT_TEST_TOTAL_POINTS / len(self.tests) * (1 - ratio))
        for arguments, expected in self.tests:
            # import sys
            # print(repr(arguments), file=sys.stderr)
            result = self.student.call(self.function_name, *arguments, context=False)
            # print(repr(self.student.exception), file=sys.stderr)
            if self.student.exception:
                all_good = False
                continue
            if assertIsInstance(result, type(expected)):
                self.points += TYPE_POINT_ADD
            else:
                all_good = False
                continue
            if self.assertEqual(result, expected):
                self.points += VALUE_POINT_ADD
            else:
                all_good = False
        if all_good:
            self.points += self.UNIT_TEST_COMPLETION_POINTS
        else:
            gently("Failing instructor unit tests")
        return all_good
